import numpy as np

from pypop7.optimizers.ds.ds import DS


class GPS(DS):
    """Generalized Pattern Search (GPS).

    .. note:: `"To converge to a local minimum, certain conditions must be met. The set of directions must
        be a positive spanning set, which means that we can construct any point using a nonnegative
        linear combination of the directions. A positive spanning set ensures that at least one of the
        directions is a descent direction from a location with a nonzero gradient."---[Kochenderfer&Wheeler, 2019]
        <https://algorithmsbook.com/optimization/files/chapter-7.pdf>`_

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma' - initial global step-size (`float`, default: `1.0`),
                * 'x'     - initial (starting) point (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'gamma' - decreasing factor of step-size (`float`, default: `0.5`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.ds.gps import GPS
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'x': 3*numpy.ones((2,)),
       ...            'sigma': 0.1,
       ...            'verbose_frequency': 500}
       >>> gps = GPS(problem, options)  # initialize the optimizer class
       >>> results = gps.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"GPS: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       GPS: 5000, 0.6182686369768672

    Attributes
    ----------
    gamma : `float`
            decreasing factor of step-size.
    sigma : `float`
            final global step-size (changed during optimization).
    x     : `array_like`
            initial (starting) point.

    References
    ----------
    Kochenderfer, M.J. and Wheeler, T.A., 2019.
    Algorithms for optimization.
    MIT Press.
    https://algorithmsbook.com/optimization/files/chapter-7.pdf
    (See Algorithm 7.6 (Page 106) for details.)

    Regis, R.G., 2016.
    On the properties of positive spanning sets and positive bases.
    Optimization and Engineering, 17(1), pp.229-262.
    https://link.springer.com/article/10.1007/s11081-015-9286-x

    Torczon, V., 1997.
    On the convergence of pattern search algorithms.
    SIAM Journal on Optimization, 7(1), pp.1-25.
    https://epubs.siam.org/doi/abs/10.1137/S1052623493250780
    """
    def __init__(self, problem, options):
        DS.__init__(self, problem, options)
        self.gamma = options.get('gamma', 0.5)  # decreasing factor of step-size (γ)

    def initialize(self, args=None, is_restart=False):
        x = self._initialize_x(is_restart)  # initial point
        y = self._evaluate_fitness(x, args)  # fitness
        # set random directions
        d = self.rng_initialization.standard_normal(size=(self.ndim_problem + 1, self.ndim_problem))
        i_d = [i for i in range(d.shape[0])]  # index of used directions
        return x, y, d, i_d

    def iterate(self, x=None, d=None, i_d=None, args=None):
        improved, best_so_far_y, fitness = False, self.best_so_far_y, []
        for i in range(d.shape[0]):
            if self._check_terminations():
                return i_d, fitness
            x = self.best_so_far_x + self.sigma*d[i_d[i]]  # opportunistic
            y = self._evaluate_fitness(x, args)
            fitness.append(y)
            if y < best_so_far_y:
                improved = True
                i_d = [i_d[i]] + i_d[:i] + i_d[(i + 1):]  # dynamic ordering
                break
        if not improved:
            self.sigma *= self.gamma  # alpha
        return i_d, fitness

    def restart_reinitialize(self, args=None, x=None, y=None, d=None, i_d=None, fitness=None):
        self._fitness_list.append(self.best_so_far_y)
        is_restart_1, is_restart_2 = self.sigma < self.sigma_threshold, False
        if len(self._fitness_list) >= self.stagnation:
            is_restart_2 = (self._fitness_list[-self.stagnation] - self._fitness_list[-1]) < self.fitness_diff
        is_restart = bool(is_restart_1) or bool(is_restart_2)
        if is_restart:
            self._print_verbose_info(fitness, y)
            self.sigma = np.copy(self._sigma_bak)
            x, y, d, i_d = self.initialize(args, is_restart)
            self._fitness_list = [self.best_so_far_y]
            self._n_generations = 0
            self._n_restart += 1
            if self.verbose:
                print(' ....... *** restart *** .......')
        return x, y, d, i_d

    def optimize(self, fitness_function=None, args=None):
        fitness = DS.optimize(self, fitness_function)
        x, y, d, i_d = self.initialize(args)
        while True:
            self._print_verbose_info(fitness, y)
            i_d, y = self.iterate(x, d, i_d, args)
            if self._check_terminations():
                break
            self._n_generations += 1
            if self.is_restart:
                x, y, d, i_d = self.restart_reinitialize(args, x, y, d, i_d, fitness)
        return self._collect(fitness, y)



"""Repeat the following paper for `POWELL`:
    Powell, M.J., 1964.
    An efficient method for finding the minimum of a function of several variables without calculating derivatives.
    The Computer Journal, 7(2), pp.155-162.
    https://academic.oup.com/comjnl/article-abstract/7/2/155/335330
    https://docs.scipy.org/doc/scipy/reference/optimize.minimize-powell.html

    Luckily our Python code could repeat the data generated by the other Python code *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import numpy as np
from scipy.optimize import minimize, Bounds

from pypop7.benchmarks.base_functions import ellipsoid, rosenbrock, rastrigin
from pypop7.optimizers.ds.powell import POWELL as Solver


if __name__ == '__main__':
    functions = [ellipsoid, rosenbrock, rastrigin]
    ndim_problem = 1000

    for f in functions:
        res = minimize(f, 4*np.ones((ndim_problem,)), method='powell',
                       bounds=Bounds(-5*np.ones((ndim_problem,)), 5*np.ones((ndim_problem,))),
                       options={'maxfev': 2e6})
        print(res)
    # ellipsoid: nfev: 20528, fun: 9.576797826010937e-26
    # rosenbrock: nfev: 817629 fun: 0.0002689337326274941
    # rastrigin: nfev: 16002 fun: 0.0

    for f in functions:
        print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -5*np.ones((ndim_problem,)),
                   'upper_boundary': 5*np.ones((ndim_problem,))}
        options = {'max_function_evaluations': 2e6,
                   'fitness_threshold': 1e-10,
                   'max_runtime': 3600,  # 1 hours
                   'seed_rng': 0,
                   'x': 4*np.ones((ndim_problem,)),
                   'sigma': 0.1,
                   'verbose': 20000,
                   'saving_fitness': 20000}
        solver = Solver(problem, options)
        results = solver.optimize()
        print(results)
        # https://github.com/Evolutionary-Intelligence/pypop/blob/main/docs/repeatability/powell/_test_powell.txt



import numpy as np


# helper function
def squeeze_and_check(x, size_gt_1=False):
    """Squeeze the input `x` into 1-d `numpy.ndarray`.
        And check whether its number of dimensions == 1. If not, raise a TypeError.
        Optionally, check whether its size > 1. If not, raise a TypeError.
    """
    x = np.squeeze(x)
    if (x.ndim == 0) and (x.size == 1):
        x = np.array([x])
    if x.ndim != 1:
        raise TypeError(f'The number of dimensions should == 1 (not {x.ndim}) after numpy.squeeze(x).')
    if size_gt_1 and not (x.size > 1):
        raise TypeError(f'The size should > 1 (not {x.size}) after numpy.squeeze(x).')
    if x.size == 0:
        raise TypeError(f'the size should != 0.')
    return x


# helper class
class BaseFunction(object):
    def __init__(self):
        pass


def sphere(x):
    y = np.sum(np.power(squeeze_and_check(x), 2))
    return y


class Sphere(BaseFunction):
    def __call__(self, x):
        return sphere(x)


def cigar(x):
    x = np.power(squeeze_and_check(x, True), 2)
    y = x[0] + (10 ** 6) * np.sum(x[1:])
    return y


class Cigar(BaseFunction):
    def __call__(self, x):
        return cigar(x)


def discus(x):  # also called tablet
    x = np.power(squeeze_and_check(x, True), 2)
    y = (10 ** 6) * x[0] + np.sum(x[1:])
    return y


class Discus(BaseFunction):  # also called Tablet
    def __call__(self, x):
        return discus(x)


def cigar_discus(x):
    x = np.power(squeeze_and_check(x, True), 2)
    if x.size == 2:
        y = x[0] + (10 ** 4) * np.sum(x) + (10 ** 6) * x[-1]
    else:
        y = x[0] + (10 ** 4) * np.sum(x[1:-1]) + (10 ** 6) * x[-1]
    return y


class CigarDiscus(BaseFunction):
    def __call__(self, x):
        return cigar_discus(x)


def ellipsoid(x):
    x = np.power(squeeze_and_check(x, True), 2)
    y = np.dot(np.power(10, 6 * np.linspace(0, 1, x.size)), x)
    return y


class Ellipsoid(BaseFunction):
    def __call__(self, x):
        return ellipsoid(x)


def different_powers(x):
    x = np.abs(squeeze_and_check(x, True))
    y = np.sum(np.power(x, 2 + 4 * np.linspace(0, 1, x.size)))
    return y


class DifferentPowers(BaseFunction):
    def __call__(self, x):
        return different_powers(x)


def schwefel221(x):
    y = np.max(np.abs(squeeze_and_check(x)))
    return y


class Schwefel221(BaseFunction):
    def __call__(self, x):
        return schwefel221(x)


def step(x):
    y = np.sum(np.power(np.floor(squeeze_and_check(x) + 0.5), 2))
    return y


class Step(BaseFunction):
    def __call__(self, x):
        return step(x)


def schwefel222(x):
    x = np.abs(squeeze_and_check(x))
    y = np.sum(x) + np.prod(x)
    return y


class Schwefel222(BaseFunction):
    def __call__(self, x):
        return schwefel222(x)


def rosenbrock(x):
    x = squeeze_and_check(x, True)
    y = 100 * np.sum(np.power(x[1:] - np.power(x[:-1], 2), 2)) + np.sum(np.power(x[:-1] - 1, 2))
    return y


class Rosenbrock(BaseFunction):
    def __call__(self, x):
        return rosenbrock(x)


def schwefel12(x):
    x = squeeze_and_check(x, True)
    x = [np.sum(x[:i + 1]) for i in range(x.size)]
    y = np.sum(np.power(x, 2))
    return y


class Schwefel12(BaseFunction):
    def __call__(self, x):
        return schwefel12(x)


def exponential(x):
    x = squeeze_and_check(x)
    y = -np.exp(-0.5 * np.sum(np.power(x, 2)))
    return y


class Exponential(BaseFunction):
    def __call__(self, x):
        return exponential(x)


def griewank(x):
    x = squeeze_and_check(x)
    y = np.sum(np.power(x, 2)) / 4000 - np.prod(np.cos(x / np.sqrt(np.arange(1, x.size + 1)))) + 1
    return y


class Griewank(BaseFunction):
    def __call__(self, x):
        return griewank(x)


def bohachevsky(x):
    x, y = squeeze_and_check(x), 0
    for i in range(x.size - 1):
        y += np.power(x[i], 2) + 2 * np.power(x[i + 1], 2) -\
             0.3 * np.cos(3 * np.pi * x[i]) - 0.4 * np.cos(4 * np.pi * x[i + 1]) + 0.7
    return y


class Bohachevsky(BaseFunction):
    def __call__(self, x):
        return bohachevsky(x)


def ackley(x):
    x = squeeze_and_check(x)
    y = -20 * np.exp(-0.2 * np.sqrt(np.sum(np.power(x, 2)) / x.size)) -\
        np.exp(np.sum(np.cos(2 * np.pi * x)) / x.size) +\
        20 + np.exp(1)
    return y


class Ackley(BaseFunction):
    def __call__(self, x):
        return ackley(x)


def rastrigin(x):
    x = squeeze_and_check(x)
    y = 10 * x.size + np.sum(np.power(x, 2) - 10 * np.cos(2 * np.pi * x))
    return y


class Rastrigin(BaseFunction):
    def __call__(self, x):
        return rastrigin(x)


def scaled_rastrigin(x):
    x, w = squeeze_and_check(x), np.power(10, np.linspace(0, 1, x.size))
    x *= w
    y = 10 * x.size + np.sum(np.power(x, 2) - 10 * np.cos(2 * np.pi * x))
    return y


class ScaledRastrigin(BaseFunction):
    def __call__(self, x):
        return scaled_rastrigin(x)


def skew_rastrigin(x):
    x = squeeze_and_check(x)
    for i in range(x.size):
        if x[i] > 0:
            x[i] *= 10
    y = rastrigin(x)
    return y


class SkewRastrigin(BaseFunction):
    def __call__(self, x):
        return skew_rastrigin(x)


def levy_montalvo(x):
    x, y = 1 + (1 / 4) * (squeeze_and_check(x) + 1), 0
    for i in range(x.size - 1):
        y += np.power(x[i] - 1, 2) * (1 + 10 * np.power(np.sin(np.pi * x[i + 1]), 2))
    y += 10 * np.power(np.sin(np.pi * x[0]), 2) + np.power(x[-1] - 1, 2)
    return (np.pi / x.size) * y


class LevyMontalvo(BaseFunction):
    def __call__(self, x):
        return levy_montalvo(x)


def michalewicz(x):
    x, y = squeeze_and_check(x), 0
    for i in range(x.size):
        y -= np.sin(x[i]) * np.power(np.sin((i + 1) * np.power(x[i], 2) / np.pi), 20)
    return y


class Michalewicz(BaseFunction):
    def __call__(self, x):
        return michalewicz(x)


def salomon(x):
    x = np.sqrt(np.sum(np.power(squeeze_and_check(x), 2)))
    y = 1 - np.cos(2 * np.pi * x) + 0.1 * x
    return y


class Salomon(BaseFunction):
    def __call__(self, x):
        return salomon(x)


def shubert(x):
    x, y = squeeze_and_check(x), 1
    for i in range(x.size):
        yy = 0
        for j in range(1, 6):
            yy += j * np.cos((j + 1) * x[i] + j)
        y *= yy
    return y


class Shubert(BaseFunction):
    def __call__(self, x):
        return shubert(x)


def schaffer(x):
    x, y = squeeze_and_check(x), 0
    for i in range(x.size - 1):
        xx = np.power(x[i], 2) + np.power(x[i + 1], 2)
        y += np.power(xx, 0.25) * (np.power(np.sin(50 * np.power(xx, 0.1)), 2) + 1)
    return y


class Schaffer(BaseFunction):
    def __call__(self, x):
        return schaffer(x)



import numpy as np

from pypop7.optimizers.cc.cc import CC


class COSYNE(CC):
    """CoOperative SYnapse NEuroevolution (COSYNE).

    .. note:: This is a wrapper of `COSYNE`, which has been implemented in the Python library `EvoTorch
       <https://docs.evotorch.ai/v0.3.0/reference/evotorch/algorithms/ga/#evotorch.algorithms.ga.Cosyne>`_,
       with slight modifications.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'          - initial global step-size for Gaussian search distribution (`float`),
                * 'n_individuals'  - number of individuals/samples, aka population size (`int`, default: `100`),
                * 'n_tournaments'  - number of tournaments for one-point crossover (`int`, default: `10`),
                * 'ratio_elitists' - ratio of elitists (`float`, default: `0.3`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.cc.cosyne import COSYNE
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'sigma': 0.3,
       ...            'x': 3*numpy.ones((2,))}
       >>> cosyne = COSYNE(problem, options)  # initialize the optimizer class
       >>> results = cosyne.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"COSYNE: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       COSYNE: 5000, 0.005023488269997175

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/yff8c6xu>`_ for more details.

    Attributes
    ----------
    n_individuals  : `int`
                     number of individuals/samples, aka population size.
    n_tournaments  : `int`
                     number of tournaments for one-point crossover.
    ratio_elitists : `float`
                     ratio of elitists.
    sigma          : `float`
                     initial global step-size for Gaussian search (mutation/sampling) distribution.

    References
    ----------
    Gomez, F., Schmidhuber, J. and Miikkulainen, R., 2008.
    Accelerated neural evolution through cooperatively coevolved synapses.
    Journal of Machine Learning Research, 9(31), pp.937-965.
    https://jmlr.org/papers/v9/gomez08a.html

    https://docs.evotorch.ai/v0.3.0/reference/evotorch/algorithms/ga/#evotorch.algorithms.ga.Cosyne
    https://github.com/nnaisense/evotorch/blob/master/src/evotorch/algorithms/ga.py
    """
    def __init__(self, problem, options):
        CC.__init__(self, problem, options)
        self.sigma = options.get('sigma')  # global step-size for Gaussian search distribution
        self.n_tournaments = options.get('n_tournament', 10)  # number of tournaments for one-point crossover
        self.ratio_elitists = options.get('ratio_elitists', 0.3)  # ratio of elitists
        self._n_elitists = int(self.ratio_elitists*self.n_individuals)  # number of elitists
        self._n_parents = int(self.n_individuals/4)  # parents for crossover and mutation

    def initialize(self, args=None, is_restart=False):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))  # population
        y = np.empty((self.n_individuals,))  # fitness
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y

    def _crossover(self, x, y):  # one-point crossover
        xx = np.empty((2*len(y), self.ndim_problem))
        for i in range(len(y)):
            left = self.rng_optimization.choice(len(y), size=(self.n_tournaments,), replace=False)
            left = x[np.argmin(y[left])]
            right = self.rng_optimization.choice(len(y), size=(self.n_tournaments,), replace=False)
            right = x[np.argmin(y[right])]
            p = self.rng_optimization.choice(self.ndim_problem)
            xx[2*i], xx[2*i + 1] = np.append(left[:p], right[p:]), np.append(right[:p], left[p:])
        return xx

    def _mutate(self, x):  # mutation for all dimensions
        x += self.sigma*self.rng_optimization.standard_normal(size=(x.shape[0], self.ndim_problem))
        x = np.clip(x, self.lower_boundary, self.upper_boundary)
        return x

    def _permute(self, x):  # different from the original paper for simplicity
        xx = np.copy(x)
        for d in range(self.ndim_problem):
            p = self.rng_optimization.choice(self.n_individuals)
            xx[:, d] = np.append(xx[p:, d], xx[:p, d])
        return xx

    def iterate(self, x=None, y=None, args=None):
        order, yy, yyy = np.argsort(y), np.empty((2*self._n_parents,)), np.empty((self.n_individuals,))
        xx = self._mutate(self._crossover(x[order[:self.n_parents]], y[order[:self.n_parents]]))
        for i in range(2*self._n_parents):
            if self._check_terminations():
                return x, y, np.append(yy, yyy)
            yy[i] = self._evaluate_fitness(xx[i], args)
        xxx = self._permute(x)
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x, y, np.append(yy, yyy)
            yyy[i] = self._evaluate_fitness(xxx[i], args)
        x = np.vstack((np.vstack((x[order[:self._n_elitists]], xx)), xxx))
        y = np.hstack((np.hstack((y[order[:self._n_elitists]], yy)), yyy))
        order = np.argsort(y)[:self.n_individuals]  # to keep population size fixed
        self._n_generations += 1
        return x[order], y[order], np.append(yy, yyy)

    def optimize(self, fitness_function=None, args=None):
        fitness = CC.optimize(self, fitness_function)
        x, y = self.initialize(args)
        yy = y  # only for printing
        while not self._check_terminations():
            self._print_verbose_info(fitness, yy)
            x, y, yy = self.iterate(x, y, args)
        return self._collect(fitness, yy)



import numpy as np

from pypop7.optimizers.es.es import ES
from pypop7.optimizers.es.r1es import R1ES


class RMES(R1ES):
    """Rank-M Evolution Strategy (RMES).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'             - initial global step-size, aka mutation strength (`float`),
                * 'mean'              - initial (starting) point, aka mean of Gaussian search distribution
                  (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'n_evolution_paths' - number of evolution paths (`int`, default: `2`),
                * 'generation_gap'    - generation gap (`int`, default: `problem['ndim_problem']`),
                * 'n_individuals'     - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_parents'         - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`),
                * 'c_cov'             - learning rate of low-rank covariance matrix (`float`, default:
                  `1.0/(3.0*np.sqrt(problem['ndim_problem']) + 5.0)`),
                * 'd_sigma'           - delay factor of cumulative step-size adaptation (`float`, default: `1.0`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.rmes import RMES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> rmes = RMES(problem, options)  # initialize the optimizer class
       >>> results = rmes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"RMES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       RMES: 5000, 5.7278132941412774e-08

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/wx3wakxj>`_ for more details.

    Attributes
    ----------
    c_cov             : `float`
                        learning rate of low-rank covariance matrix adaptation.
    d_sigma           : `float`
                        delay factor of cumulative step-size adaptation.
    generation_gap    : `int`
                        generation gap.
    mean              : `array_like`
                        initial (starting) point, aka mean of Gaussian search distribution.
    n_evolution_paths : `int`
                        number of evolution paths.
    n_individuals     : `int`
                        number of offspring, aka offspring population size.
    n_parents         : `int`
                        number of parents, aka parental population size.
    sigma             : `float`
                        final global step-size, aka mutation strength.

    References
    ----------
    Li, Z. and Zhang, Q., 2018.
    A simple yet efficient evolution strategy for large-scale black-box optimization.
    IEEE Transactions on Evolutionary Computation, 22(5), pp.637-646.
    https://ieeexplore.ieee.org/abstract/document/8080257
    """
    def __init__(self, problem, options):
        R1ES.__init__(self, problem, options)
        self.n_evolution_paths = options.get('n_evolution_paths', 2)
        self.generation_gap = options.get('generation_gap', self.ndim_problem)
        self._a = np.sqrt(1.0 - self.c_cov)
        self._a_m = np.power(self._a, self.n_evolution_paths)
        self._b = np.sqrt(self.c_cov)

    def initialize(self, args=None, is_restart=False):
        x, mean, p, s, y = R1ES.initialize(self, args, is_restart)
        mp = np.zeros((self.n_evolution_paths, self.ndim_problem))  # multiple evolution paths
        t_hat = np.zeros((self.n_evolution_paths,))
        return x, mean, p, s, mp, t_hat, y

    def iterate(self, x=None, mean=None, mp=None, y=None, args=None):
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            z = self.rng_optimization.standard_normal((self.ndim_problem,))
            sum_p = np.zeros((self.ndim_problem,))
            for i in np.arange(self.n_evolution_paths) + 1:
                r = self.rng_optimization.standard_normal()
                sum_p += np.power(self._a, self.n_evolution_paths - i)*r*mp[i - 1]
            x[k] = mean + self.sigma*(self._a_m*z + self._b*sum_p)
            y[k] = self._evaluate_fitness(x[k], args)
        return x, y

    def _update_distribution(self, x=None, mean=None, p=None, s=None,
                             mp=None, t_hat=None, y=None, y_bak=None):
        mean, p, s = R1ES._update_distribution(self, x, mean, p, s, y, y_bak)
        # update multiple evolution paths
        t_min = np.min(np.diff(t_hat))
        i_apostrophe = np.argmin(np.diff(t_hat))
        i_apostrophe += 1
        if (t_min > self.generation_gap) or (self._n_generations < self.n_evolution_paths):
            i_apostrophe = 0
        for i in range(i_apostrophe, self.n_evolution_paths - 1):
            mp[i], t_hat[i] = mp[i + 1], t_hat[i + 1]
        mp[-1], t_hat[-1] = p, self._n_generations
        return mean, p, s, mp, t_hat

    def restart_reinitialize(self, args=None, x=None, mean=None, p=None, s=None,
                             mp=None, t_hat=None, y=None, fitness=None):
        if self.is_restart and ES.restart_reinitialize(self, y):
            x, mean, p, s, mp, t_hat, y = self.initialize(args, True)
            self._print_verbose_info(fitness, y[0])
            self.d_sigma *= 2.0
        return x, mean, p, s, mp, t_hat, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, p, s, mp, t_hat, y = self.initialize(args)
        self._print_verbose_info(fitness, y[0])
        while not self.termination_signal:
            y_bak = np.copy(y)
            # sample and evaluate offspring population
            x, y = self.iterate(x, mean, mp, y, args)
            self._n_generations += 1
            self._print_verbose_info(fitness, y)
            if self._check_terminations():
                break
            mean, p, s, mp, t_hat = self._update_distribution(x, mean, p, s, mp, t_hat, y, y_bak)
            x, mean, p, s, mp, t_hat, y = self.restart_reinitialize(
                args, x, mean, p, s, mp, t_hat, y, fitness)
        results = self._collect(fitness, y, mean)
        results['p'] = p
        results['s'] = s
        return results



"""Repeat the following paper for `DCEM`:
    Amos, B. and Yarats, D., 2020, November.
    The differentiable cross-entropy method.
    In International Conference on Machine Learning (pp. 291-302). PMLR.
    http://proceedings.mlr.press/v119/amos20a.html

    Luckily our Python code could repeat the data generated by the original Python code *well*.
    Therefore, we argue that its repeatability could be **well-documented**.



    You can run the following Python script (note that first download the library from
      https://github.com/facebookresearch/dcem):
    ----------------------------------------------------------------------------------
    import numpy as np
    import torch

    from dcem import dcem


    def f(x):
        w = np.power(10, 6 * np.linspace(0, 1, x.shape[2]))
        x = x.detach().numpy()
        y = np.empty((1, x.shape[1]))
        for i in range(x.shape[1]):
            y[0, i] = np.dot(w, np.power(x[0, i], 2))
        y = torch.from_numpy(y)
        return y


    mu = np.random.default_rng(0).uniform(-5*np.ones((1, 100)), 5*np.ones((1, 100)))
    opt = dcem(f=f, nx=100, init_mu=torch.from_numpy(mu), init_sigma=1.0,
               n_sample=1000, n_elite=200, n_iter=int(2e6/1000), temp=1.0, iter_eps=0)
    print(f(opt.unsqueeze(1)))
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import ellipsoid
from pypop7.optimizers.cem.dcem import DCEM as Solver


if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 100
    problem = {'fitness_function': ellipsoid,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5*np.ones((ndim_problem,)),
               'upper_boundary': 5*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 2e6,
               'seed_rng': 0,
               'sigma': 1.0,
               'verbose': 20,
               'saving_fitness': 2000}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(ellipsoid(results['mean']))  # 3.568666713353976e-08 vs 3.6024e-08 (from the original code)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))



import numpy as np
from scipy.stats import cauchy

from pypop7.optimizers.de.jade import JADE


class SHADE(JADE):
    """Success-History based Adaptive Differential Evolution (SHADE).

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default: `100`),
                * 'mu'            - mean of normal distribution for adaptation of crossover probability (`float`,
                  default: `0.5`),
                * 'median'        - median of Cauchy distribution for adaptation of mutation factor (`float`,
                  default: `0.5`),
                *  'h'            - length of historical memory (`int`, default: `100`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.de.shade import SHADE
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 0}
       >>> shade = SHADE(problem, options)  # initialize the optimizer class
       >>> results = shade.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"SHADE: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       SHADE: 5000, 6.231767087114823e-05

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/vm3w7se4>`_ for more details.

    Attributes
    ----------
    h             : `int`
                    length of historical memory.
    median        : `float`
                    median of Cauchy distribution for adaptation of mutation factor.
    mu            : `float`
                    mean of normal distribution for adaptation of crossover probability.
    n_individuals : `int`
                    number of offspring, aka offspring population size.

    References
    ----------
    Tanabe, R. and Fukunaga, A., 2013, June.
    Success-history based parameter adaptation for differential evolution.
    In IEEE Congress on Evolutionary Computation (pp. 71-78). IEEE.
    https://ieeexplore.ieee.org/document/6557555
    """
    def __init__(self, problem, options):
        JADE.__init__(self, problem, options)
        self.h = options.get('h', 100)  # length of historical memory
        assert 0 < self.h
        self.m_mu = np.ones(self.h)*self.mu  # means of normal distribution
        self.m_median = np.ones(self.h)*self.median  # medians of Cauchy distribution
        self._k = 0  # index to update
        self.p_min = 2.0/self.n_individuals

    def mutate(self, x=None, y=None, a=None):
        x_mu = np.empty((self.n_individuals, self.ndim_problem))  # mutated population
        f_mu = np.empty((self.n_individuals,))  # mutated mutation factors
        x_un = np.vstack((np.copy(x), a))  # union of population x and archive a
        r = self.rng_optimization.choice(self.h, (self.n_individuals,))
        order = np.argsort(y)[:]
        p = (0.2 - self.p_min)*self.rng_optimization.random((self.n_individuals,)) + self.p_min
        idx = [order[self.rng_optimization.choice(int(i))] for i in np.ceil(p*self.n_individuals)]
        for k in range(self.n_individuals):
            f_mu[k] = cauchy.rvs(loc=self.m_median[r[k]], scale=0.1, random_state=self.rng_optimization)
            while f_mu[k] <= 0.0:
                f_mu[k] = cauchy.rvs(loc=self.m_median[r[k]], scale=0.1, random_state=self.rng_optimization)
            if f_mu[k] > 1.0:
                f_mu[k] = 1.0
            r1 = self.rng_optimization.choice([i for i in range(self.n_individuals) if i != k])
            r2 = self.rng_optimization.choice([i for i in range(len(x_un)) if i != k and i != r1])
            x_mu[k] = x[k] + f_mu[k]*(x[idx[k]] - x[k]) + f_mu[k]*(x[r1] - x_un[r2])
        return x_mu, f_mu, r

    def crossover(self, x_mu=None, x=None, r=None):
        x_cr = np.copy(x)
        p_cr = np.empty((self.n_individuals,))  # crossover probabilities
        for k in range(self.n_individuals):
            p_cr[k] = self.rng_optimization.normal(self.m_mu[r[k]], 0.1)
            p_cr[k] = np.minimum(np.maximum(p_cr[k], 0.0), 1.0)
            i_rand = self.rng_optimization.integers(self.ndim_problem)
            for i in range(self.ndim_problem):
                if (i == i_rand) or (self.rng_optimization.random() < p_cr[k]):
                    x_cr[k, i] = x_mu[k, i]
        return x_cr, p_cr

    def select(self, args=None, x=None, y=None, x_cr=None, a=None, f_mu=None, p_cr=None):
        # set successful mutation factors, crossover probabilities, fitness differences
        f, p, d = np.empty((0,)), np.empty((0,)), np.empty((0,))
        for k in range(self.n_individuals):
            if self._check_terminations():
                break
            yy = self._evaluate_fitness(x_cr[k], args)
            if yy < y[k]:
                a = np.vstack((a, x[k]))  # archive of inferior solutions
                f = np.hstack((f, f_mu[k]))  # archive of successful mutation factors
                p = np.hstack((p, p_cr[k]))  # archive of successful crossover probabilities
                d = np.hstack((d, y[k] - yy))  # archive of successful fitness differences
                x[k], y[k] = x_cr[k], yy
        if (len(p) != 0) and (len(f) != 0):
            w = d/np.sum(d)
            self.m_mu[self._k] = np.sum(w*p)  # for normal distribution
            self.m_median[self._k] = np.sum(w*np.power(f, 2))/np.sum(w*f)  # for Cauchy distribution
            self._k = (self._k + 1) % self.h
        return x, y, a

    def iterate(self, x=None, y=None, a=None, args=None):
        x_mu, f_mu, r = self.mutate(x, y, a)
        x_cr, p_cr = self.crossover(x_mu, x, r)
        x_cr = self.bound(x_cr, x)
        x, y, a = self.select(args, x, y, x_cr, a, f_mu, p_cr)
        if len(a) > self.n_individuals:  # randomly remove solutions to keep archive size fixed
            a = np.delete(a, self.rng_optimization.choice(len(a), (len(a) - self.n_individuals,), False), 0)
        self._n_generations += 1
        return x, y, a



from pypop7.optimizers.cc.cc import CC  # abstract class for all CC classes
from pypop7.optimizers.cc.coea import COEA
from pypop7.optimizers.cc.hcc import HCC
from pypop7.optimizers.cc.cosyne import COSYNE
from pypop7.optimizers.cc.cocma import COCMA


__all__ = [CC,  # Cooperative Coevolution (CC)
           COEA,  # CoOperative co-Evolutionary Algorithm (COEA)
           HCC,  # Hierarchical Cooperative Co-evolution (HCC)
           COSYNE,  # CoOperative SYnapse NEuroevolution (COSYNE)
           COCMA]  # CoOperative CO-evolutionary Covariance Matrix Adaptation (COCMA)



"""Repeat the following paper for `LEP`:
    Lee, C.Y. and Yao, X., 2004.
    Evolutionary programming using mutations based on the Lévy probability distribution.
    IEEE Transactions on Evolutionary Computation, 8(1), pp.1-13.
    https://ieeexplore.ieee.org/document/1266370

    Luckily our Python code could repeat the data reported in the original paper *nearly well*.
    Therefore, we argue that its repeatability could be **well-documented** (*at least partly*).
"""
import numpy as np

from pypop7.benchmarks.base_functions import sphere, rosenbrock, rastrigin, ackley
from pypop7.optimizers.ep.lep import LEP


if __name__ == '__main__':
    ndim_problem = 30

    problem = {'fitness_function': ackley,
               'ndim_problem': ndim_problem,
               'lower_boundary': -32*np.ones((ndim_problem,)),
               'upper_boundary': 32*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500*100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    lep = LEP(problem, options)
    results = lep.optimize()
    print(results['best_so_far_y'])
    # 0.09633170691380899 vs 0.974767

    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100*np.ones((ndim_problem,)),
               'upper_boundary': 100*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500*100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    lep = LEP(problem, options)
    results = lep.optimize()
    print(results['best_so_far_y'])
    # 0.006244546982883647 vs 0.001979

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -30*np.ones((ndim_problem,)),
               'upper_boundary': 30*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500*100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    lep = LEP(problem, options)
    results = lep.optimize()
    print(results['best_so_far_y'])
    # 37.60584627168424 vs 72.343559

    problem = {'fitness_function': rastrigin,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5.12*np.ones((ndim_problem,)),
               'upper_boundary': 5.12*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500*100,
               'seed_rng': 0,  # undefined in the original paper
               'sigma': 3.0}
    lep = LEP(problem, options)
    results = lep.optimize()
    print(results['best_so_far_y'])
    # 425.4823087914274 vs 38.266239



import numpy as np  # engine for numerical computing

from pypop7.optimizers.eda.eda import EDA


class RPEDA(EDA):
    """Random-Projection Estimation of Distribution Algorithm (RPEDA).

    .. note:: `RPEDA` uses **random matrix theory (RMT)** to sample individuals on multiple embedded subspaces,
       though it still evaluates all individuals on the original search space. It has a **quadractic** time
       complexity w.r.t. each sampling for large-scale black-box optimization.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - number of offspring, offspring population size (`int`, default: `300`),
                * 'n_parents'     - number of parents, parental population size (`int`, default:
                  `int(0.25*options['n_individuals'])`),
                * 'k'             - projection dimensionality (`int`, default: `3`),
                * 'm'             - number of random projection matrices (`int`, default:
                  `int(np.ceil(4*options['n_individuals']/options['k'])`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy  # engine for numerical computing
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.eda.rpeda import RPEDA
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 20,
       ...            'lower_boundary': -5*numpy.ones((20,)),
       ...            'upper_boundary': 5*numpy.ones((20,))}
       >>> options = {'max_function_evaluations': 500000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'k': 2}
       >>> rpeda = RPEDA(problem, options)  # initialize the optimizer class
       >>> results = rpeda.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"RPEDA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       RPEDA: 500000, 15.67048345324486

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    n_parents     : `int`
                    number of parents, aka parental population size.
    k             : `int`
                    projection dimensionality.
    m             : `int`
                    number of random projection matrices.

    References
    ----------
    Kabán, A., Bootkrajang, J. and Durrant, R.J., 2016.
    Toward large-scale continuous EDA: A random matrix theory perspective.
    Evolutionary Computation, 24(2), pp.255-291.
    https://direct.mit.edu/evco/article-abstract/24/2/255/1016/Toward-Large-Scale-Continuous-EDA-A-Random-Matrix
    """
    def __init__(self, problem, options):
        EDA.__init__(self, problem, options)
        self.n_individuals = options.get('n_individuals', 300)  # population size
        n_parents = int(0.25*self.n_individuals)
        self.n_parents = options.get('n_parents', n_parents)  # number of selected individuals
        self.k = options.get('k', 3)  # projection dimensionality
        assert self.k < self.ndim_problem
        size_rpm = int(np.ceil(4*self.ndim_problem/self.k))
        self.m = options.get('m', size_rpm)  # number of random projection matrices
        self._sq_1_d = np.sqrt(1.0/self.ndim_problem)
        self._sq_nsk = np.sqrt(self.ndim_problem*self.m/self.k)

    def initialize(self):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))  # population
        xx = np.copy(x)  # top individuals
        y = np.empty((self.n_individuals,))  # fitness
        return x, xx, y

    def iterate(self, xx=None, y=None, args=None):
        mean = np.mean(xx, axis=0)
        diff = xx - mean  # centred points to be projected
        x = np.zeros((self.n_individuals, self.ndim_problem))
        for i in range(self.m):
            # project into a k-dimensional subspace
            r = self.rng_optimization.standard_normal((self.ndim_problem, self.k))*self._sq_1_d
            pm = np.dot(diff, r)  # projection
            cm = np.cov(np.transpose(pm))  # covariance matrix
            samples = self.rng_optimization.multivariate_normal(np.zeros((self.k,)), cm,
                                                                size=(self.n_individuals,))
            # recover into the original space
            x += np.dot(samples, np.transpose(r))
        x /= self.m  # averaging
        x *= self._sq_nsk  # scaling
        for i in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            if i == 0:  # elitism
                x[0] = xx[0]
            else:
                x[i] = np.clip(x[i] + mean, self.lower_boundary, self.upper_boundary)
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y

    def optimize(self, fitness_function=None, args=None):
        fitness = EDA.optimize(self, fitness_function)
        x, xx, y = self.initialize()
        while not self.termination_signal:
            x, y = self.iterate(xx, y, args)
            if self._check_terminations():
                break
            order = np.argsort(y)[:self.n_parents]  # to select top individuals
            xx = np.copy(x[order])
            self._n_generations += 1
            self._print_verbose_info(fitness, y)
        return self._collect(fitness, y)



import numpy as np

from pypop7.optimizers.es.es import ES


class CMAES(ES):
    """Covariance Matrix Adaptation Evolution Strategy (CMAES).

    .. note:: `CMAES` is widely recognized as one of the **State Of The Art (SOTA)** for black-box optimization,
       according to the latest `Nature <https://www.nature.com/articles/nature14544>`_ review of Evolutionary
       Computation.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.cmaes import CMAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'is_restart': False,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> cmaes = CMAES(problem, options)  # initialize the optimizer class
       >>> results = cmaes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"CMAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       CMAES: 5000, 9.11305771685916e-09

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/4mysrjwe>`_ for more details.

    Attributes
    ----------
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search distribution.
    n_individuals : `int`
                    number of offspring, aka offspring population size / sample size.
    n_parents     : `int`
                    number of parents, aka parental population size / number of positively selected search points.
    sigma         : `float`
                    final global step-size, aka mutation strength.

    References
    ----------
    Hansen, N., 2023.
    The CMA evolution strategy: A tutorial.
    arXiv preprint arXiv:1604.00772.
    https://arxiv.org/abs/1604.00772

    Hansen, N., Müller, S.D. and Koumoutsakos, P., 2003.
    Reducing the time complexity of the derandomized evolution strategy with covariance matrix adaptation (CMA-ES).
    Evolutionary Computation, 11(1), pp.1-18.
    https://direct.mit.edu/evco/article-abstract/11/1/1/1139/Reducing-the-Time-Complexity-of-the-Derandomized

    Hansen, N. and Ostermeier, A., 2001.
    Completely derandomized self-adaptation in evolution strategies.
    Evolutionary Computation, 9(2), pp.159-195.
    https://direct.mit.edu/evco/article-abstract/9/2/159/892/Completely-Derandomized-Self-Adaptation-in

    Hansen, N. and Ostermeier, A., 1996, May.
    Adapting arbitrary normal mutation distributions in evolution strategies: The covariance matrix adaptation.
    In Proceedings of IEEE International Conference on Evolutionary Computation (pp. 312-317). IEEE.
    https://ieeexplore.ieee.org/abstract/document/542381

    See the lightweight implementation of CMA-ES from cyberagent.ai:
    https://github.com/CyberAgentAILab/cmaes
    """
    def __init__(self, problem, options):
        self.options = options
        ES.__init__(self, problem, options)
        assert self.n_individuals >= 2
        self._mu_eff = None  # effective sample size of selected samples / variance effective selection mass 
        self._mu_eff_minus = None
        self._w = None
        self._alpha_cov = 2.0
        self.c_s = None
        self.d_sigma = None
        self.c_c = None  # decay rate for cumulation (evolution) path for rank-one update of CMA
        self.c_1 = None  # learning rate for rank-one update of CMA
        self.c_w = None
        self._p_s_1 = None  # for evolution path update of cumulative step-length adaptation (CSA)
        self._p_s_2 = None
        self._p_c_1 = None  # for evolution path update of CMA
        self._p_c_2 = None
        self.n_generations, self._n_generations = None, None

    def _set_c_c(self):  # to set decay rate for cumulation (evolution) path for rank-one update of CMA
        return (4.0 + self._mu_eff/self.ndim_problem)/(self.ndim_problem + 4.0 + 2.0*self._mu_eff/self.ndim_problem)

    def _set_c_w(self):
        # minus 1e-8 for large population size (according to https://github.com/CyberAgentAILab/cmaes)
        return np.minimum(1.0 - self.c_1 - 1e-8, self._alpha_cov*(self._mu_eff + 1.0/self._mu_eff - 2.0) /
                          (np.power(self.ndim_problem + 2.0, 2) + self._alpha_cov*self._mu_eff/2.0))

    def _set_d_sigma(self):
        return 1.0 + self.c_s + 2.0*np.maximum(0.0, np.sqrt((self._mu_eff - 1.0)/(self.ndim_problem + 1.0)) - 1.0)

    def initialize(self, is_restart=False):
        w_apostrophe = np.log((self.n_individuals + 1.0)/2.0) - np.log(np.arange(self.n_individuals) + 1.0)
        self._mu_eff = np.power(np.sum(w_apostrophe[:self.n_parents]), 2)/np.sum(
            np.power(w_apostrophe[:self.n_parents], 2))
        self._mu_eff_minus = np.power(np.sum(w_apostrophe[self.n_parents:]), 2)/np.sum(
            np.power(w_apostrophe[self.n_parents:], 2))
        self.c_s = self.options.get('c_s', (self._mu_eff + 2.0)/(self._mu_eff + self.ndim_problem + 5.0))
        self.d_sigma = self.options.get('d_sigma', self._set_d_sigma())
        self.c_c = self.options.get('c_c', self._set_c_c())
        self.c_1 = self.options.get('c_1', self._alpha_cov/(np.square(self.ndim_problem + 1.3) + self._mu_eff))
        self.c_w = self.options.get('c_w', self._set_c_w())
        w_min = np.min([1.0 + self.c_1/self.c_w, 1.0 + 2.0*self._mu_eff_minus/(self._mu_eff + 2.0),
                        (1.0 - self.c_1 - self.c_w)/(self.ndim_problem*self.c_w)])
        self._w = np.where(w_apostrophe >= 0, 1.0/np.sum(w_apostrophe[w_apostrophe > 0])*w_apostrophe,
                           w_min/(-np.sum(w_apostrophe[w_apostrophe < 0]))*w_apostrophe)
        self._p_s_1 = 1.0 - self.c_s
        self._p_s_2 = np.sqrt(self._mu_eff*self.c_s*(2.0 - self.c_s))
        self._p_c_1 = 1.0 - self.c_c
        self._p_c_2 = np.sqrt(self._mu_eff*self.c_c*(2.0 - self.c_c))
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        p_s = np.zeros((self.ndim_problem,))  # evolution path for cumulative step-length adaptation (CSA)
        p_c = np.zeros((self.ndim_problem,))  # evolution path for covariance matrix adaptation (CMA)
        cm = np.eye(self.ndim_problem)  # covariance matrix of Gaussian search distribution
        eig_ve = np.eye(self.ndim_problem)  # eigenvectors of covariance matrix
        eig_va = np.ones((self.ndim_problem,))  # eigenvalues of covariance matrix
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        self._list_initial_mean.append(np.copy(mean))
        self._n_generations = self.n_generations = 0
        return x, mean, p_s, p_c, cm, eig_ve, eig_va, y

    def iterate(self, x=None, mean=None, eig_ve=None, eig_va=None, y=None, args=None):
        for k in range(self.n_individuals):  # to sample offspring population
            if self._check_terminations():
                return x, y
            z = self.rng_optimization.standard_normal((self.ndim_problem,))  # Gaussian noise for mutation
            x[k] = mean + self.sigma*np.dot(np.dot(eig_ve, np.diag(eig_va)), z)  # offspring individual
            y[k] = self._evaluate_fitness(x[k], args)  # fitness
        return x, y

    def update_distribution(self, x=None, mean=None, p_s=None, p_c=None, cm=None, eig_ve=None, eig_va=None, y=None):
        order = np.argsort(y)
        d = (x - mean)/self.sigma
        wd = np.dot(self._w[:self.n_parents], d[order[:self.n_parents]])
        # update mean via weighted recombination
        mean += self.sigma*wd
        # update global step-size (CSA)
        cm_minus_half = np.dot(np.dot(eig_ve, np.diag(1.0/eig_va)), np.transpose(eig_ve))
        p_s = self._p_s_1*p_s + self._p_s_2*np.dot(cm_minus_half, wd)
        self.sigma *= np.exp(self.c_s/self.d_sigma*(np.linalg.norm(p_s)/self._e_chi - 1.0))
        # update covariance matrix (CMA)
        h_s = 1.0 if np.linalg.norm(p_s)/np.sqrt(1.0 - np.power(1.0 - self.c_s, 2*(self.n_generations + 1))) < (
                1.4 + 2.0/(self.ndim_problem + 1.0))*self._e_chi else 0.0
        p_c = self._p_c_1*p_c + h_s*self._p_c_2*wd
        w_o = self._w*np.where(self._w >= 0, 1.0, self.ndim_problem/(np.power(np.linalg.norm(
            np.dot(cm_minus_half, np.transpose(d)), axis=0), 2) + 1e-8))
        cm = ((1.0 + self.c_1*(1.0 - h_s)*self.c_c*(2.0 - self.c_c) - self.c_1 - self.c_w*np.sum(self._w))*cm +
              self.c_1*np.outer(p_c, p_c))  # rank-one update
        for i in range(self.n_individuals):  # rank-μ update (to estimate variances of sampled *steps*)
            cm += self.c_w*w_o[i]*np.outer(d[order[i]], d[order[i]])
        # do eigen decomposition (SVD)
        cm = (cm + np.transpose(cm))/2.0
        eig_va, eig_ve = np.linalg.eigh(cm)
        eig_va = np.sqrt(np.where(eig_va < 0, 1e-8, eig_va))
        cm = np.dot(np.dot(eig_ve, np.diag(np.power(eig_va, 2))), np.transpose(eig_ve))
        return mean, p_s, p_c, cm, eig_ve, eig_va

    def restart_reinitialize(self, x=None, mean=None, p_s=None, p_c=None,
                             cm=None, eig_ve=None, eig_va=None, y=None):
        if ES.restart_reinitialize(self, y):
            x, mean, p_s, p_c, cm, eig_ve, eig_va, y = self.initialize(True)
        return x, mean, p_s, p_c, cm, eig_ve, eig_va, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        x, mean, p_s, p_c, cm, eig_ve, eig_va, y = self.initialize()
        while True:
            # sample and evaluate offspring population
            x, y = self.iterate(x, mean, eig_ve, eig_va, y, args)
            if self._check_terminations():
                break
            self._print_verbose_info(fitness, y)
            self.n_generations += 1
            self._n_generations = self.n_generations
            mean, p_s, p_c, cm, eig_ve, eig_va = self.update_distribution(x, mean, p_s, p_c, cm, eig_ve, eig_va, y)
            if self.is_restart:
                x, mean, p_s, p_c, cm, eig_ve, eig_va, y = self.restart_reinitialize(
                    x, mean, p_s, p_c, cm, eig_ve, eig_va, y)
        results = self._collect(fitness, y, mean)
        results['p_s'] = p_s
        results['p_c'] = p_c
        results['eig_va'] = eig_va
        # by default, do NOT save covariance matrix of search distribution in order to save memory,
        # owing to its *quadratic* space complexity
        return results



import time

import numpy as np

from pypop7.optimizers.es.lmcma import LMCMA  # for upper-level
from pypop7.optimizers.es.cmaes import CMAES  # for lower-level
from pypop7.optimizers.cc import CC


class HCC(CC):
    """Hierarchical Cooperative Co-evolution (HCC).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular setting (`key`):
                * 'n_individuals'   - number of individuals/samples, aka population size (`int`, default: `100`).
                * 'sigma'           - initial global step-size (`float`, default:
                  `problem['upper_boundary'] - problem['lower_boundary']/3.0`),
                * 'ndim_subproblem' - dimensionality of subproblem for decomposition (`int`, default: `30`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.cc.hcc import HCC
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'x': 3*numpy.ones((2,))}
       >>> hcc = HCC(problem, options)  # initialize the optimizer class
       >>> results = hcc.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"HCC: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       HCC: 5000, 0.0006576371426980234

    For its correctness checking of coding, we cannot provide the code-based repeatability report, since this
    implementation combines two different papers. To our knowledge, few well-designed open-source code of `CC` is
    available for non-separable black-box optimization.

    Attributes
    ----------
    n_individuals   : `int`
                      number of individuals/samples, aka population size.
    sigma           : `float`
                      initial global step-size.
    ndim_subproblem : `int`
                      dimensionality of subproblem for decomposition.

    References
    ----------
    Mei, Y., Omidvar, M.N., Li, X. and Yao, X., 2016.
    A competitive divide-and-conquer algorithm for unconstrained large-scale black-box optimization.
    ACM Transactions on Mathematical Software, 42(2), pp.1-24.
    https://dl.acm.org/doi/10.1145/2791291

    Gomez, F.J. and Schmidhuber, J., 2005, June.
    Co-evolving recurrent neurons learn deep memory POMDPs.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 491-498). ACM.
    https://dl.acm.org/doi/10.1145/1068009.1068092
    """
    def __init__(self, problem, options):
        CC.__init__(self, problem, options)
        self.sigma = options.get('sigma')  # global step size
        assert self.sigma is None or self.sigma > 0.0
        self.ndim_subproblem = int(options.get('ndim_subproblem', 30))
        assert self.ndim_subproblem > 0

    def initialize(self, arg=None):
        # for lower-level CMA-ES
        self.best_so_far_x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary)
        self.best_so_far_y = self._evaluate_fitness(self.best_so_far_x, arg)
        sub_optimizers = []
        for i in range(int(np.ceil(self.ndim_problem/self.ndim_subproblem))):
            ii = range(i*self.ndim_subproblem, np.minimum((i + 1)*self.ndim_subproblem, self.ndim_problem))
            problem = {'ndim_problem': len(ii),  # cyclic decomposition
                       'lower_boundary': self.lower_boundary[ii],
                       'upper_boundary': self.upper_boundary[ii]}
            if self.sigma is None:
                sigma = np.min((self.upper_boundary[ii] - self.lower_boundary[ii])/3.0)
            else:
                sigma = self.sigma
            options = {'seed_rng': self.rng_initialization.integers(np.iinfo(np.int64).max),
                       'sigma': sigma,
                       'max_runtime': self.max_runtime,
                       'fitness_threshold': self.fitness_threshold,
                       'verbose': False}
            cma = CMAES(problem, options)
            cma.start_time = time.time()
            sub_optimizers.append(cma)
        # for upper-level LM-CMA
        problem = {'ndim_problem': self.ndim_problem,
                   'lower_boundary': self.lower_boundary,
                   'upper_boundary': self.upper_boundary}
        if self.sigma is None:
            sigma = np.min((self.upper_boundary - self.lower_boundary)/3.0)
        else:
            sigma = self.sigma
        options = {'seed_rng': self.rng_initialization.integers(np.iinfo(np.int64).max),
                   'sigma': sigma,
                   'max_runtime': self.max_runtime,
                   'fitness_threshold': self.fitness_threshold,
                   'verbose': False}
        lmcma = LMCMA(problem, options)
        return sub_optimizers, self.best_so_far_y, lmcma

    def optimize(self, fitness_function=None, args=None):
        fitness, is_initialization = CC.optimize(self, fitness_function), True
        sub_optimizers, y, lmcma = self.initialize(args)
        # run upper-level LM-CMA
        lm_mean, lm_x, lm_p_c, lm_s, lm_vm, lm_pm, lm_b, lm_d, lm_y = lmcma.initialize()  # for upper-level LM-CMA
        lmcma.start_time = time.time()
        lmcma.fitness_function = self._evaluate_fitness
        # run lower-level CMA-ES
        x_s, mean_s, ps_s, pc_s, cm_s, ee_s, ea_s, y_s = [], [], [], [], [], [], [], []  # for lower-level CMA-ES
        while not self._check_terminations():
            self._print_verbose_info(fitness, y)
            y = []
            # run upper-level LM-CMA
            lm_y_bak = np.copy(lm_y)
            lmcma.max_function_evaluations = (lmcma.n_function_evaluations +
                                              self.max_function_evaluations - self.n_function_evaluations)
            lm_x, lm_y = lmcma.iterate(lm_mean, lm_x, lm_pm, lm_vm, lm_y, lm_b, args)
            lm_mean, lm_p_c, lm_s, lm_vm, lm_pm, lm_b, lm_d = lmcma.update_distribution(
                lm_mean, lm_x, lm_p_c, lm_s, lm_vm, lm_pm, lm_b, lm_d, lm_y, lm_y_bak)
            y.extend(lm_y)
            # run lower-level CMA-ES
            if is_initialization:
                is_initialization = False
                for i, opt in enumerate(sub_optimizers):
                    if self._check_terminations():
                        break

                    x, mean, p_s, p_c, cm, eig_ve, eig_va, yy = opt.initialize()
                    x_s.append(x)
                    mean_s.append(mean)
                    ps_s.append(p_s)
                    pc_s.append(p_c)
                    cm_s.append(cm)
                    ee_s.append(eig_ve)
                    ea_s.append(eig_va)
                    y_s.append(yy)
            else:
                for i, opt in enumerate(sub_optimizers):
                    ii = range(i*self.ndim_subproblem, np.minimum((i + 1)*self.ndim_subproblem, self.ndim_problem))

                    def sub_function(sub_x):  # to define sub-function for each sub-optimizer
                        best_so_far_x = np.copy(self.best_so_far_x)
                        best_so_far_x[ii] = sub_x
                        return self._evaluate_fitness(best_so_far_x, args)
                    opt.fitness_function = sub_function
                    opt.max_function_evaluations = (opt.n_function_evaluations +
                                                    self.max_function_evaluations - self.n_function_evaluations)
                    x_s[i], y_s[i] = opt.iterate(x_s[i], mean_s[i], ee_s[i], ea_s[i], y_s[i], args)
                    y.extend(y_s[i])
                    if self._check_terminations():
                        break
                    opt.n_generations += 1
                    mean_s[i], ps_s[i], pc_s[i], cm_s[i], ee_s[i], ea_s[i] = opt.update_distribution(
                        x_s[i], mean_s[i], ps_s[i], pc_s[i], cm_s[i], ee_s[i], ea_s[i], y_s[i])
                if self.best_so_far_y < lmcma.best_so_far_y:  # to communicate between upper and lower levels
                    lm_mean = np.copy(self.best_so_far_x)
            self._n_generations += 1
        return self._collect(fitness, y)



"""Repeat the following paper for `SAMAES`:
    Beyer, H.G., 2020, July.
    Design principles for matrix adaptation evolution strategies.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation Companion (pp. 682-700). ACM.
    https://dl.acm.org/doi/abs/10.1145/3377929.3389870

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/blob/main/docs/repeatability/samaes/_repeat_samaes.png

    Luckily our Python code could repeat the data reported in the paper *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.optimizers.es.samaes import SAMAES


# set parameters for lens shape optimization
weight = 0.9  # weight of focus function
r = 7  # radius of lens
h = 1  # trapezoidal slices of height
b = 20  # distance between lens and object
eps = 1.5  # refraction index
d_init, sigma = 3.0, 1.0  # initialization


# define objective function to be minimized
def f_lens(x):  # refer to [Beyer, 2020, ACM-GECCO] for the mathematical details
    n = len(x)
    focus = r - ((h*np.arange(1, n) - 0.5) + b/h*(eps - 1)*np.transpose(np.abs(x[1:]) - np.abs(x[:(n-1)])))
    mass = h*(np.sum(np.abs(x[1:(n-1)])) + 0.5*(np.abs(x[0]) + np.abs(x[n-1])))
    return weight*np.sum(focus**2) + (1.0 - weight)*mass


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    ndim_problem = 15  # dimension of objective function
    problem = {'fitness_function': f_lens,  # objective function
               'ndim_problem': ndim_problem}  # number of dimensionality of objective function
    options = {'max_function_evaluations': 1000*20,  # maximum of function evaluations
               'seed_rng': 1,  # seed of random number generation (for repeatability)
               'n_individuals': 20,
               'n_parents': 5,
               'x': d_init*np.ones((ndim_problem,)),  # initial mean of Gaussian search distribution
               'sigma': sigma,  # global step-size of Gaussian search distribution (not necessarily an optimal value)
               'saving_fitness': 1,  # to record best-so-far fitness every 50 function evaluations
               'is_restart': False}  # whether or not to run the (default) restart process
    results = SAMAES(problem, options).optimize()
    print(results)
    plt.plot(results['fitness'][:, 0]/20, results['fitness'][:, 1], 'b')
    plt.xlabel('g')
    plt.xticks([0, 200, 400, 600, 800, 1000])
    plt.xlim([0, 1000])
    plt.ylabel('f_lens')
    plt.yscale('log')
    plt.yticks([1e0, 1e1, 1e2, 1e3])
    plt.show()



import numpy as np

from pypop7.optimizers.pso.pso import PSO


class CLPSO(PSO):
    """Comprehensive Learning Particle Swarm Optimizer (CLPSO).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - swarm (population) size, aka number of particles (`int`, default: `20`),
                * 'c'             - comprehensive learning rate (`float`, default: `1.49445`),
                * 'm'             - refreshing gap (`int`, default: `7`),
                * 'max_ratio_v'   - maximal ratio of velocities w.r.t. search range (`float`, default: `0.2`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.pso.clpso import CLPSO
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> clpso = CLPSO(problem, options)  # initialize the optimizer class
       >>> results = clpso.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"CLPSO: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       CLPSO: 5000, 7.184727085112434e-05

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/f3pp4nfh>`_ for more details.

    Attributes
    ----------
    c             : `float`
                    comprehensive learning rate.
    m             : `int`
                    refreshing gap.
    max_ratio_v   : `float`
                    maximal ratio of velocities w.r.t. search range.
    n_individuals : `int`
                    swarm (population) size, aka number of particles.

    References
    ----------
    Liang, J.J., Qin, A.K., Suganthan, P.N. and Baskar, S., 2006.
    Comprehensive learning particle swarm optimizer for global optimization of multimodal functions.
    IEEE Transactions on Evolutionary Computation, 10(3), pp.281-295.
    https://ieeexplore.ieee.org/abstract/document/1637688

    See the original MATLAB source code from Prof. Suganthan:
    https://github.com/P-N-Suganthan/CODES/blob/master/2006-IEEE-TEC-CLPSO.zip
    """
    def __init__(self, problem, options):
        PSO.__init__(self, problem, options)
        self.c = options.get('c', 1.49445)  # comprehensive learning rate
        assert self.c > 0.0
        self.m = options.get('m', 7)  # refreshing gap
        assert self.m > 0
        pc = 5.0*np.linspace(0, 1, self.n_individuals)
        self._pc = 0.5*(np.exp(pc) - np.exp(pc[0]))/(np.exp(pc[-1]) - np.exp(pc[0]))
        # set number of successive generations each particle has not improved its best fitness
        self._flag = np.zeros((self.n_individuals,))
        # set linearly decreasing inertia weights from 0.9 to 0.2
        self._w = 0.9 - 0.7*(np.arange(self._max_generations) + 1.0)/self._max_generations

    def _learn_topology(self, p_x, p_y, i, n_x):
        if self._flag[i] >= self.m:  # if no improved for successive generations
            self._flag[i], exemplars = 0, i*np.ones((self.ndim_problem,))
            for d in range(self.ndim_problem):
                if self.rng_optimization.random() < self._pc[i]:  # learn from other
                    # use tournament selection
                    left, right = self.rng_optimization.choice(self.n_individuals, 2, replace=False)
                    if p_y[left] < p_y[right]:
                        n_x[i, d], exemplars[d] = p_x[left, d], left
                    else:
                        n_x[i, d], exemplars[d] = p_x[right, d], right
                else:  # inherit from its own best position
                    n_x[i, d] = p_x[i, d]
            if np.alltrue(exemplars == i):  # learning from other when all exemplars are itself
                ndim = self.rng_optimization.integers(self.ndim_problem)  # randomly selected dimension
                exemplar = self.rng_optimization.choice([k for k in range(self.n_individuals) if k != i])
                n_x[i, ndim] = p_x[exemplar, ndim]

    def iterate(self, v=None, x=None, y=None, p_x=None, p_y=None, n_x=None, args=None):
        for i in range(self.n_individuals):  # online (rather batch) update
            if self._check_terminations():
                return v, x, y, p_x, p_y, n_x
            self._learn_topology(p_x, p_y, i, n_x)
            v[i] = (self._w[min(self._n_generations, len(self._w) - 1)]*v[i] + self.c*self.rng_optimization.uniform(
                size=(self.ndim_problem,))*(n_x[i] - x[i]))  # velocity update
            v[i] = np.clip(v[i], self._min_v, self._max_v)
            x[i] += v[i]  # position update
            if self.is_bound:
                x[i] = np.clip(x[i], self.lower_boundary, self.upper_boundary)
            y[i] = self._evaluate_fitness(x[i], args)
            if y[i] < p_y[i]:  # personally-best position update
                p_x[i], p_y[i] = np.clip(x[i], self.lower_boundary, self.upper_boundary), y[i]
            else:
                self._flag[i] += 1
        self._n_generations += 1
        return v, x, y, p_x, p_y, n_x



"""Repeat the following paper for `RPEDA`:
    Kabán, A., Bootkrajang, J. and Durrant, R.J., 2016.
    Toward large-scale continuous EDA: A random matrix theory perspective.
    Evolutionary Computation, 24(2), pp.255-291.
    https://direct.mit.edu/evco/article-abstract/24/2/255/1016/Toward-Large-Scale-Continuous-EDA-A-Random-Matrix

    Luckily our Python code could repeat the data reported in the original paper *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock
from pypop7.optimizers.eda.rpeda import RPEDA


if __name__ == '__main__':
    ndim_problem = 1000
    for f in [rosenbrock]:
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -100*np.ones((ndim_problem,)),
                   'upper_boundary': 100*np.ones((ndim_problem,))}
        options = {'max_function_evaluations': 6e5,
                   'm': 1000,
                   'seed_rng': 0,
                   'verbose': 100,
                   'saving_fitness': 1}
        rpeda = RPEDA(problem, options)
        results = rpeda.optimize()
        print(f.__name__, results['best_so_far_y'])
        # rosenbrock 989.8908871376332 vs mean 1614.7 std 249.44 (from the original paper)



"""Repeat the following paper for `ENES`:
    Wierstra, D., Schaul, T., Glasmachers, T., Sun, Y., Peters, J. and Schmidhuber, J., 2014.
    Natural evolution strategies.
    Journal of Machine Learning Research, 15(1), pp.949-980.
    https://jmlr.org/papers/v15/wierstra14a.html
    https://github.com/pybrain/pybrain/blob/master/pybrain/optimization/distributionbased/nes.py



    import numpy as np
    from pybrain.optimization.distributionbased.nes import ExactNES as ENES


    def sphere(x):  # to be maximized
        return -np.sum(np.power(x, 2))


    np.random.seed(5)
    solver = ENES(sphere, 4*np.ones((10,)), maxEvaluations=5e5, verbose=True, importanceMixing=False)
    solver.x = 4*np.ones((10,))
    solver.learn()
    # ('Evals:', 43100)
    # ('Step:', 430, 'best:', -2.2205949442220285e-20)
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import sphere
from pypop7.optimizers.nes.enes import ENES as Solver


if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 10
    for f in [sphere]:
        print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem}
        options = {'max_function_evaluations': 43100,
                   'seed_rng': 0,
                   'x': 4*np.ones((ndim_problem,)),
                   'saving_fitness': 1,
                   'is_restart': False}
        solver = Solver(problem, options)
        results = solver.optimize()
        print(results)
        print(results['best_so_far_y'])  # 2.295365297463382e-20
        print('*** Runtime: {:7.5e}'.format(time.time() - start_run))



"""Repeat the following paper for `JADE`:
    Zhang, J., and Sanderson, A. C. 2009.
    JADE: Adaptive differential evolution with optional external archive.
    IEEE Transactions on Evolutionary Computation, 13(5), 945–958.
    https://ieeexplore.ieee.org/document/5208221/

    Luckily our Python code could repeat the data reported in the original paper *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import numpy as np

from pypop7.benchmarks.base_functions import sphere, step, rosenbrock, rastrigin, ackley
from pypop7.optimizers.de.jade import JADE


if __name__ == '__main__':
    ndim_problem = 30

    problem = {'fitness_function': sphere,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100*np.ones((ndim_problem,)),
               'upper_boundary': 100*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500*100,
               'n_individuals': 100,
               'seed_rng': 0,  # undefined in the original paper
               }
    jade = JADE(problem, options)
    results = jade.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 1.2487346869372509e-64 vs 1.3e-54 (from the original paper)

    problem = {'fitness_function': step,
               'ndim_problem': ndim_problem,
               'lower_boundary': -100*np.ones((ndim_problem,)),
               'upper_boundary': 100*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 1500*100,
               'n_individuals': 100,
               'seed_rng': 0,  # undefined in the original paper
               }
    jade = JADE(problem, options)
    results = jade.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.0 vs 0 (from the original paper)

    problem = {'fitness_function': rosenbrock,
               'ndim_problem': ndim_problem,
               'lower_boundary': -30*np.ones((ndim_problem,)),
               'upper_boundary': 30*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 3000*100,
               'n_individuals': 100,
               'seed_rng': 0,  # undefined in the original paper
               }
    jade = JADE(problem, options)
    results = jade.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.0 vs 3.2e-01 (from the original paper)

    problem = {'fitness_function': rastrigin,
               'ndim_problem': ndim_problem,
               'lower_boundary': -5.12*np.ones((ndim_problem,)),
               'upper_boundary': 5.12*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 5000*100,
               'n_individuals': 100,
               'seed_rng': 0,  # undefined in the original paper
               }
    jade = JADE(problem, options)
    results = jade.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 0.0 vs 0 (from the original paper)

    problem = {'fitness_function': ackley,
               'ndim_problem': ndim_problem,
               'lower_boundary': -32*np.ones((ndim_problem,)),
               'upper_boundary': 32*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 2000*100,
               'n_individuals': 100,
               'seed_rng': 0,  # undefined in the original paper
               }
    jade = JADE(problem, options)
    results = jade.optimize()
    print(results)
    print(results['best_so_far_y'])
    # 3.9968028886505635e-15 vs 4.4e-15 (from the original paper)



import numpy as np

from pypop7.optimizers.ga.ga import GA


class GL25(GA):
    """Global and Local genetic algorithm (GL25).

    .. note:: `25` means that 25 percentage of function evaluations (or runtime) are first used for *global* search
       while the remaining 75 percentage are then used for *local* search.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'alpha'           - global step-size for crossover (`float`, default: `0.8`),
                * 'n_female_global' - number of female at global search stage (`int`, default: `200`),
                * 'n_male_global'   - number of male at global search stage (`int`, default: `400`),
                * 'n_female_local'  - number of female at local search stage (`int`, default: `5`),
                * 'n_male_local'    - number of male at local search stage (`int`, default: `100`),
                * 'p_global'        - percentage of global search stage (`float`, default: `0.25`),.

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.ga.gl25 import GL25
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> gl25 = GL25(problem, options)  # initialize the optimizer class
       >>> results = gl25.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"GL25: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       GL25: 5000, 1.0505276479694516e-05

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/ytzffmbc>`_ for more details.

    Attributes
    ----------
    alpha           : `float`
                      global step-size for crossover.
    n_female_global : `int`
                      number of female at global search stage.
    n_female_local  : `int`
                      number of female at local search stage.
    n_individuals   : `int`
                      population size.
    n_male_global   : `int`
                      number of male at global search stage.
    n_male_local    : `int`
                      number of male at local search stage.
    p_global        : `float`
                      percentage of global search stage.

    References
    ----------
    García-Martínez, C., Lozano, M., Herrera, F., Molina, D. and Sánchez, A.M., 2008.
    Global and local real-coded genetic algorithms based on parent-centric crossover operators.
    European Journal of Operational Research, 185(3), pp.1088-1113.
    https://www.sciencedirect.com/science/article/abs/pii/S0377221706006308
    """
    def __init__(self, problem, options):
        GA.__init__(self, problem, options)
        self.alpha = options.get('alpha', 0.8)
        assert self.alpha > 0.0
        self.p_global = options.get('p_global', 0.25)  # percentage of global search stage
        assert 0.0 <= self.p_global <= 1.0
        self.n_female_global = options.get('n_female_global', 200)  # number of female at global search stage
        assert self.n_female_global > 0
        self.n_male_global = options.get('n_male_global', 400)  # number of male at global search stage
        assert self.n_male_global > 0
        self.n_female_local = options.get('n_female_local', 5)  # number of female at local search stage
        assert self.n_female_local > 0
        self.n_male_local = options.get('n_male_local', 100)  # number of male at local search stage
        assert self.n_male_local > 0
        self.n_individuals = int(np.maximum(self.n_male_global, self.n_male_local))
        self._assortative_mating = 5
        self._n_selected = np.zeros((self.n_individuals,))  # number of individuals selected as female
        # set maximum of function evaluations and runtime for global search stage
        self._max_fe_global = self.p_global*self.max_function_evaluations
        self._max_runtime_global = self.p_global*self.max_runtime

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))  # population
        y = np.empty((self.n_individuals,))  # fitness
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y

    def iterate(self, x=None, y=None, n_female=None, n_male=None, args=None):
        order = np.argsort(y)
        x_male, female = x[order[range(n_male)]], order[range(n_female)]
        x_female, _n_selected = x[female], self._n_selected[female]
        # use the uniform fertility selection (UFS) as female selection mechanism
        female = np.argmin(_n_selected)
        _n_selected[female] += 1
        self._n_selected[order[range(n_female)]] = _n_selected
        female = x_female[female]
        # use negative assortative mating (NAM) as male selection mechanism
        distances = np.empty((self._assortative_mating,))
        male = self.rng_optimization.choice(n_male, size=self._assortative_mating, replace=False)
        for i, m in enumerate(male):
            distances[i] = np.linalg.norm(female - x_male[m])
        male = x_male[male[np.argmax(distances)]]
        # use the parent-centric BLX crossover operator
        interval = np.abs(female - male)
        l, u = female - interval*self.alpha, female + interval*self.alpha
        xx = self.rng_optimization.uniform(np.clip(l, self.lower_boundary, self.upper_boundary),
                                           np.clip(u, self.lower_boundary, self.upper_boundary))
        yy = self._evaluate_fitness(xx, args)
        # use the replace worst (RW) strategy
        if yy < y[order[-1]]:
            x[order[-1]], y[order[-1]], self._n_selected[order[-1]] = xx, yy, 0
        self._n_generations += 1
        return x, yy

    def optimize(self, fitness_function=None, args=None):
        fitness, is_switch = GA.optimize(self, fitness_function), True
        x, y = self.initialize(args)
        yy = y  # only for printing
        while not self._check_terminations():
            self._print_verbose_info(fitness, yy)
            if self.n_function_evaluations >= self._max_fe_global or self.runtime >= self._max_runtime_global:
                if is_switch:  # local search
                    init, is_switch = range(np.maximum(self.n_female_local, self.n_male_local)), False
                    x, y, self._n_selected = x[init], y[init], self._n_selected[init]
                x, yy = self.iterate(x, y, self.n_female_local, self.n_male_local, args)
            else:  # global search
                x, yy = self.iterate(x, y, self.n_female_global, self.n_male_global, args)
        return self._collect(fitness, yy)



"""Repeat the following paper for `FCMAES`:
    Li, Z., Zhang, Q., Lin, X. and Zhen, H.L., 2018.
    Fast covariance matrix adaptation for large-scale black-box optimization.
    IEEE Transactions on Cybernetics, 50(5), pp.2073-2083.
    https://ieeexplore.ieee.org/abstract/document/8533604

    All generated figures can be accessed via the following link:
    https://github.com/Evolutionary-Intelligence/pypop/tree/main/docs/repeatability/fcmaes

    Luckily our Python code could repeat the data reported in the paper *well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import pickle

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

from pypop7.benchmarks.base_functions import cigar, ellipsoid, discus, different_powers, rosenbrock
from pypop7.optimizers.es.fcmaes import FCMAES


def plot(function, ndim):
    plt.figure()
    result = pickle.load(open(function + '_' + str(ndim) + '.pickle', 'rb'))
    plt.yscale('log')
    plt.plot(result['fitness'][:, 0], result['fitness'][:, 1], color='r')
    plt.ylim([1e-10, 1e10])
    plt.yticks(ticks=[1e-10, 1e-5, 1e0, 1e5, 1e10])
    if function == 'cigar':
        plt.xlim([0e5, 5e5])
        plt.xticks([0e5, 1e5, 2e5, 3e5, 4e5, 5e5])
    elif function == 'ellipsoid':
        plt.xlim([0e7, 2.5e7])
        plt.xticks([0e7, 0.5e7, 1e7, 1.5e7, 2e7, 2.5e7])
    elif function == 'discus':
        plt.xlim([0e7, 5e7])
        plt.xticks([0e7, 1e7, 2e7, 3e7, 4e7, 5e7])
    elif function == 'different_powers':
        plt.xlim([0e6, 10e6])
        plt.xticks([0e6, 2e6, 4e6, 6e6, 8e6, 10e6])
    elif function == 'rosenbrock':
        plt.xscale('log')
        plt.xlim([1e2, 1e8])
        plt.xticks([1e2, 1e4, 1e6, 1e8])
    plt.xlabel("evaluations")
    plt.ylabel("objective value")
    plt.title(function.capitalize())
    plt.show()


if __name__ == '__main__':
    sns.set_theme(style='darkgrid')
    ndim_problem = 1024
    for f in [rosenbrock, cigar, ellipsoid, discus, different_powers]:
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -10*np.ones((ndim_problem,)),
                   'upper_boundary': 10*np.ones((ndim_problem,))}
        options = {'fitness_threshold': 1e-8,
                   'seed_rng': 1,  # not given in the original paper
                   'sigma': 20.0/3.0,
                   'is_restart': False,
                   'saving_fitness': 1}
        rmes = FCMAES(problem, options)
        results = rmes.optimize()
        print(results)
        pickle.dump(results, open(f.__name__ + '_' + str(ndim_problem) + '.pickle', 'wb'))
        plot(f.__name__, ndim_problem)



import numpy as np

from pypop7.optimizers.ga.ga import GA


class GENITOR(GA):
    """GENetic ImplemenTOR (GENITOR).

    .. note:: `"Selective pressure and population diversity should be controlled as directly as possible."---[Whitley,
       1989] <https://dl.acm.org/doi/10.5555/93126.93169>`_

       This is a *slightly modified* version of `GENITOR` for continuous optimization. Originally `GENITOR` was proposed
       to solve challenging `neuroevolution <https://www.nature.com/articles/s42256-018-0006-z>`_ problems by Whitley,
       `the recipient of IEEE Evolutionary Computation Pioneer Award 2022 <https://tinyurl.com/456as566>`_.

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular setting (`key`):
                * 'n_individuals'  - population size (`int`, default: `100`),
                * 'cv_prob'        - crossover probability (`float`, default: `0.5`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.ga.genitor import GENITOR
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> genitor = GENITOR(problem, options)  # initialize the optimizer class
       >>> results = genitor.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"GENITOR: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       GENITOR: 5000, 0.004382445279905116

    For its correctness checking of coding, the code-based repeatability report cannot be provided owing to
    the lack of its simulation environment.

    Attributes
    ----------
    cv_prob       : `float`
                    crossover probability.
    n_individuals : `int`
                    population size.

    References
    ----------
    https://www.cs.colostate.edu/~genitor/

    Whitley, D., Dominic, S., Das, R. and Anderson, C.W., 1993.
    Genetic reinforcement learning for neurocontrol problems.
    Machine Learning, 13, pp.259-284.
    https://link.springer.com/article/10.1023/A:1022674030396

    Whitley, D., 1989, December.
    The GENITOR algorithm and selection pressure: Why rank-based allocation of reproductive trials is best.
    In Proceedings of International Conference on Genetic Algorithms (pp. 116-121).
    https://dl.acm.org/doi/10.5555/93126.93169
    """
    def __init__(self, problem, options):
        GA.__init__(self, problem, options)
        self.cv_prob = options.get('cv_prob', 0.5)  # crossover probability
        assert 0.0 <= self.cv_prob <= 1.0
        _rank_prob = np.arange(self.n_individuals, 0, -1) - 1.0
        self.rank_prob = _rank_prob/np.sum(_rank_prob)

    def initialize(self, args=None):
        x = self.rng_initialization.uniform(self.initial_lower_boundary, self.initial_upper_boundary,
                                            size=(self.n_individuals, self.ndim_problem))  # population
        y = np.empty((self.n_individuals,))  # fitness
        cv_probs = self.cv_prob*np.ones((self.n_individuals,))
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y, cv_probs

    def iterate(self, x=None, y=None, cv_probs=None, args=None):
        order, xx, yy = np.argsort(y), None, None
        # use rank-based selection for two parents

        offspring = self.rng_optimization.choice(order, size=2, replace=False, p=self.rank_prob)
        if self.rng_optimization.random() < cv_probs[offspring[0]]:  # crossover
            # use intermediate crossover (not one-point crossover proposed in the original paper)
            xx = (x[offspring[0]] + x[offspring[1]])/2.0
            yy = self._evaluate_fitness(xx, args)
            if yy < y[order[-1]]:  # to replace the worst individual
                x[order[-1]], y[order[-1]] = xx, yy
                cv_probs[offspring[0]] += 0.1
            else:
                cv_probs[offspring[0]] -= 0.1
            cv_probs[offspring[0]] = np.maximum(0.05, np.minimum(0.95, cv_probs[offspring[0]]))
        else:  # mutation
            xx = np.copy(x[offspring[0]])  # offspring
            xx += self.rng_optimization.uniform(self.lower_boundary, self.upper_boundary)/10.0
            yy = self._evaluate_fitness(xx, args)
            if yy < y[order[-1]]:  # to replace the worst individual
                x[order[-1]], y[order[-1]] = xx, yy
        self._n_generations += 1
        return x, yy, cv_probs

    def optimize(self, fitness_function=None, args=None):
        fitness = GA.optimize(self, fitness_function)
        x, y, cv_probs = self.initialize(args)
        yy = y  # only for printing
        while not self._check_terminations():
            self._print_verbose_info(fitness, yy)
            x, yy, cv_probs = self.iterate(x, y, cv_probs, args)
        return self._collect(fitness, yy)



import numpy as np

from pypop7.optimizers.es.es import ES


class EMA(object):  # exponential moving average
    def __init__(self, ndim):
        self.m = np.zeros((ndim,))

    def update(self, d):
        self.m = d


class VKDCMA(ES):
    """Projection-based Covariance Matrix Adaptation (VKDCMA).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.vkdcma import VKDCMA
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> vkdcma = VKDCMA(problem, options)  # initialize the optimizer class
       >>> results = vkdcma.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"VKDCMA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       VKDCMA: 5000, 1.7960753151742513e-13

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/mrxvx4hk>`_ for more details.

    Attributes
    ----------
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search distribution.
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    n_parents     : `int`
                    number of parents, aka parental population size.
    sigma         : `float`
                    final global step-size, aka mutation strength.

    References
    ----------
    Akimoto, Y. and Hansen, N., 2016, September.
    Online model selection for restricted covariance matrix adaptation.
    In Parallel Problem Solving from Nature. Springer International Publishing.
    https://link.springer.com/chapter/10.1007/978-3-319-45823-6_1

    Akimoto, Y. and Hansen, N., 2016, July.
    Projection-based restricted covariance matrix adaptation for high dimension.
    In Proceedings of Annual Genetic and Evolutionary Computation Conference 2016 (pp. 197-204). ACM.
    https://dl.acm.org/doi/abs/10.1145/2908812.2908863

    See the official Python version from Prof. Akimoto:
    https://gist.github.com/youheiakimoto/2fb26c0ace43c22b8f19c7796e69e108
    """
    def _get_m(self, d_x, d, v, s):
        dx_d = d_x/d
        v_dxd = np.dot(v[:self.k_a], dx_d)
        return np.sum(dx_d*dx_d) - np.sum((v_dxd*v_dxd)*(s[:self.k_a]/(s[:self.k_a] + 1.0)))

    def _get_lr(self, k):
        c_1 = 2.0/(self.ndim_problem*(k + 1.0) + self.ndim_problem + 2.0*(k + 2.0) + self._mu_eff)
        c_mu = min(1.0 - c_1, 2.0*(self._mu_eff - 2.0 + 1.0/self._mu_eff)/(
                self.ndim_problem*(k + 1) + 4.0*(k + 2.0) + self._mu_eff))
        return c_1, c_mu, np.sqrt(c_1)

    def _get_det(self, d, s):
        return 2.0*np.sum(np.log(d)) + np.sum(np.log(1.0 + s[:self.k_a]))

    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.options = options
        self.k, self.k_a, self.k_i = 0, 0, 0
        self.c_s, self.d_s = 0.3, np.sqrt(self.ndim_problem) 
        self._injection = False
        self.c_1, self.c_mu, self.c_c = None, None, None
        self.l_s, self.l_d, self.l_c = None, None, None
        self.e_s, self.e_d, self.e_c = None, None, None
    
    def initialize(self, is_restart=False):
        self.k, self.k_a, self.k_i = 0, 0, 0
        self.l_s = np.log(self.sigma)
        self.l_d = 2.0*np.log(np.ones((self.ndim_problem,)))
        self.l_c = np.zeros((self.ndim_problem,))
        self.e_s = EMA(1)
        self.e_d = EMA(self.ndim_problem)
        self.e_c = EMA(self.ndim_problem)
        self.c_1, self.c_mu, self.c_c = self._get_lr(self.k)
        self.c_s, self.d_s = 0.3, np.sqrt(self.ndim_problem) 
        self._injection = False
        d = np.ones((self.ndim_problem,))
        p_s = 0
        v = np.zeros((self.k, self.ndim_problem))
        s = np.zeros((self.ndim_problem,))
        p_c = np.zeros((self.ndim_problem,))
        d_x = np.zeros((self.ndim_problem,))
        u = np.zeros((self.ndim_problem, self.k + self.n_parents + 1))
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        x = np.zeros((self.n_individuals, self.ndim_problem))
        y = np.zeros((self.n_individuals,))
        return d, p_s, v, s, p_c, d_x, u, mean, x, y

    def iterate(self, d=None, p_s=None, v=None, s=None, p_c=None, d_x=None,
                u=None, mean=None, x=None, y=None, args=None):
        k, k_a = self.k, self.k_a
        z = (self.rng_optimization.standard_normal((self.n_individuals, self.ndim_problem)) + np.dot(
            self.rng_optimization.standard_normal((self.n_individuals, k_a))*np.sqrt(
                s[:k_a]), v[:k_a]))*d
        if self._injection:
            zz = (np.linalg.norm(self.rng_optimization.standard_normal((self.ndim_problem,)))/np.sqrt(
                self._get_m(d_x, d, v, s)))*d_x
            z[0], z[1] = zz, -zz
        x = mean + self.sigma*z
        for i in range(self.n_individuals):
            if self._check_terminations():
                return d, p_s, v, s, p_c, d_x, u, mean, x, y
            y[i] = self._evaluate_fitness(x[i], args)
        order = np.argsort(y)
        s_z = z[order[:self.n_parents]]
        d_x = np.dot(self._w, s_z)
        mean += self.sigma*d_x
        if self._injection:
            p_s += self.c_s*((np.where(order == 1)[0][0] - np.where(order == 0)[0][0])/(
                self.n_individuals - 1.0) - p_s)
            self.sigma *= np.exp(p_s/self.d_s)
            h_s = p_s < 0.5
        else:
            self._injection, h_s = True, True
        p_c = (1.0 - self.c_c)*p_c + h_s*np.sqrt(self.c_c*(2.0 - self.c_c)*self._mu_eff)*d_x
        if self.c_mu == 0.0:
            r_u = k_a + 1
            alpha = np.sqrt(abs(1.0 - self.c_mu - self.c_1 + self.c_1*(
                1.0 - h_s)*self.c_c*(2.0 - self.c_c)))
            u[:, :k_a] = (v[:k_a].T*(np.sqrt(s[:k_a])*alpha))
            u[:, r_u - 1] = np.sqrt(self.c_1)*(p_c/d)
        elif self.c_1 == 0.0:
            r_u = k_a + self.n_parents
            alpha = np.sqrt(abs(1.0 - self.c_mu - self.c_1 + self.c_1*(
                1.0 - h_s)*self.c_c*(2.0 - self.c_c)))
            u[:, :k_a] = (v[:k_a].T*(np.sqrt(s[:k_a])*alpha))
            u[:, k_a:r_u] = np.sqrt(self.c_mu)*np.sqrt(self._w)*(s_z/d).T
        else:
            r_u = k_a + self.n_parents + 1
            alpha = np.sqrt(abs(1.0 - self.c_mu - self.c_1 + self.c_1*(
                1.0 - h_s)*self.c_c*(2.0 - self.c_c)))
            u[:, :k_a] = (v[:k_a].T*(np.sqrt(s[:k_a])*alpha))
            u[:, k_a:r_u - 1] = np.sqrt(self.c_mu)*np.sqrt(self._w)*(s_z/d).T
            u[:, r_u - 1] = np.sqrt(self.c_1)*(p_c/d)
        if self.ndim_problem > r_u:
            dd, right = np.linalg.eigh(np.dot(u[:, :r_u].T, u[:, :r_u]))
            i_e = np.argsort(dd)[::-1]
            gamma = 0 if r_u <= k else dd[i_e[k:]].sum()/(self.ndim_problem - k)
            self.k_a = k_a = min(int(np.sum(dd >= 0)), k)
            s[:k_a] = (dd[i_e[:k_a]] - gamma)/(alpha*alpha + gamma)
            v[:k_a] = (np.dot(u[:, :r_u], right[:, i_e[:k_a]])/np.sqrt(dd[i_e[:k_a]])).T
        else:
            dd, left = np.linalg.eigh(np.dot(u[:, :r_u], u[:, :r_u].T))
            i_e = np.argsort(dd)[::-1]
            gamma = 0 if r_u <= k else dd[i_e[k:]].sum()/(self.ndim_problem - k)
            self.k_a = k_a = min(int(np.sum(dd >= 0)), k)
            s[:k_a] = (dd[i_e[:k_a]] - gamma)/(alpha*alpha + gamma)
            v[:k_a] = left[:, i_e[:k_a]].T
        d *= np.sqrt((alpha*alpha + np.sum(u[:, :r_u]*u[:, :r_u], axis=1))/(
            1.0 + np.dot(s[:k_a], v[:k_a]*v[:k_a])))
        e = np.exp(self._get_det(d, s)/self.ndim_problem/2.0)
        d, p_c = d/e, p_c/e
        self.k_i += 1
        self.e_s.update(np.log(self.sigma) - self.l_s)
        l_s_c = self.e_s.m/(0.5*min(1.0, self.n_individuals/self.ndim_problem)/3.0)
        self.l_s = np.log(self.sigma)
        self.e_d.update(2.0*np.log(d) + np.log(1.0 + np.dot(s[:self.k], v[:self.k]**2)) - self.l_d)
        l_d_c = self.e_d.m/(self.c_mu + self.c_1)
        self.l_d = 2.0*np.log(d) + np.log(1.0 + np.dot(s[:self.k], v[:self.k]**2))
        self.e_c.update(np.log(1.0 + s) - self.l_c)
        l_l_c = self.e_c.m/(self.c_mu + self.c_1)
        self.l_c = np.log(1.0 + s)
        k_i = (self.k_i > (2.0*self.ndim_problem - 1.0))*(self.k < (self.ndim_problem - 1.0))*np.all(
            (1.0 + s[:self.k]) > 30.0)*(np.abs(l_s_c) < 0.1)*np.all(np.abs(l_d_c) < 1.0)
        k_d = (self.k > 0)*(1.0 + s[:self.k] < 30.0)*(l_l_c[:self.k] < 0.0)
        if (self.k_i > (2*self.ndim_problem - 1)) and k_i:
            self.k_a = k
            self.k = new_k = min(max(int(np.ceil(self.k*1.414)), self.k + 1), (self.ndim_problem - 1))
            v = np.vstack((v, np.zeros((new_k - k, self.ndim_problem))))
            u = np.empty((self.ndim_problem, new_k + self.n_parents + 1))
            (self.c_1, self.c_mu, self.c_c) = self._get_lr(self.k)
            self.k_i = 0
        elif self.k_i > k*(2*self.ndim_problem - 1) and np.any(k_d):
            keep = np.logical_not(k_d)
            new_k = max(np.count_nonzero(keep), 0)
            v = v[keep]
            s[:new_k] = (s[:keep.shape[0]])[keep]
            s[new_k:] = 0
            self.k = self.k_a = new_k
            (self.c_1, self.c_mu, self.c_c) = self._get_lr(self.k)
        e = np.exp(self._get_det(d, s)/self.ndim_problem/2.0)
        d, p_c = d/e, p_c/e
        return d, p_s, v, s, p_c, d_x, u, mean, x, y

    def restart_reinitialize(self, d=None, p_s=None, v=None, s=None, p_c=None,
                             d_x=None, u=None, mean=None, x=None, y=None):
        if self.is_restart and ES.restart_reinitialize(self, y):
            d, p_s, v, s, p_c, d_x, u, mean, x, y = self.initialize(True)
        return d, p_s, v, s, p_c, d_x, u, mean, x, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        d, p_s, v, s, p_c, d_x, u, mean, x, y = self.initialize()
        while not self.termination_signal:
            d, p_s, v, s, p_c, d_x, u, mean, x, y = self.iterate(d, p_s, v, s, p_c, d_x, u, mean, x, y, args)
            if self._check_terminations():
                break
            self._print_verbose_info(fitness, y)
            self._n_generations += 1
            d, p_s, v, s, p_c, d_x, u, mean, x, y = self.restart_reinitialize(d, p_s, v, s, p_c, d_x, u, mean, x, y)
        results = self._collect(fitness, y, mean)
        results['d'] = d
        return results



"""Repeat the following paper for `HJ`:
    Hooke, R. and Jeeves, T.A., 1961.
    “Direct search” solution of numerical and statistical problems.
    Journal of the ACM, 8(2), pp.212-229.
    https://dl.acm.org/doi/10.1145/321062.321069

    Luckily our Python code could repeat the data generated by the other Python code *well*.
    Therefore, we argue that its repeatability could be **well-documented**.



    The Python reference script is given below (note that first install `pymoo` via `pip install pymoo`):
    -----------------------------------------------------------------------------------------------------
    from pymoo.algorithms.soo.nonconvex.pattern import PatternSearch
    from pymoo.problems.single import Ackley
    from pymoo.optimize import minimize

    problem = Ackley(n_var=100)
    algorithm = PatternSearch(init_delta=0.1)
    res = minimize(problem=problem, algorithm=algorithm, termination=('n_eval', 1e6),  verbose=True, seed=0)
    print(res)
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import ackley
from pypop7.optimizers.ds.hj import HJ as Solver


if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 100
    for f in [ackley]:
        print('*' * 7 + ' ' + f.__name__ + ' ' + '*' * 7)
        problem = {'fitness_function': f,
                   'ndim_problem': ndim_problem,
                   'lower_boundary': -32.768*np.ones((ndim_problem,)),
                   'upper_boundary': 32.768*np.ones((ndim_problem,))}
        options = {'max_function_evaluations': 1e6,
                   'seed_rng': 0,
                   'sigma': 0.1,
                   'is_restart': False}
        solver = Solver(problem, options)
        results = solver.optimize()
        print(results)  # 19.544448155486872 vs 1.924563E+01 (from `pymoo`)
        print('*** Runtime: {:7.5e}'.format(time.time() - start_run))



import numpy as np

from pypop7.optimizers.es.es import ES


class LMCMAES(ES):
    """Limited-Memory Covariance Matrix Adaptation Evolution Strategy (LMCMAES).

    .. note:: For perhaps better performance, please use its lateset version called `LMCMA
       <https://tinyurl.com/mry5dw36>`_. Here we include it mainly for *benchmarking* purpose.

    Parameters
    ----------
    problem : `dict`
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : `dict`
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime to be allowed (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'sigma'         - initial global step-size, aka mutation strength (`float`),
                * 'mean'          - initial (starting) point, aka mean of Gaussian search distribution (`array_like`),

                  * if not given, it will draw a random sample from the uniform distribution whose search range is
                    bounded by `problem['lower_boundary']` and `problem['upper_boundary']`.

                * 'm'             - number of direction vectors (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_steps'       - target number of generations between vectors (`int`, default: `options['m']`),
                * 'c_c'           - learning rate for evolution path update (`float`, default: `1.0/options['m']`).
                * 'c_1'           - learning rate for covariance matrix adaptation (`float`, default:
                  `1.0/(10.0*np.log(problem['ndim_problem'] + 1.0))`),
                * 'c_s'           - learning rate for population success rule (`float`, default: `0.3`),
                * 'd_s'           - delay rate for population success rule (`float`, default: `1.0`),
                * 'z_star'        - target success rate for population success rule (`float`, default: `0.25`),
                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default:
                  `4 + int(3*np.log(problem['ndim_problem']))`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`).

    Examples
    --------
    Use the optimizer to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.es.lmcmaes import LMCMAES
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5*numpy.ones((2,)),
       ...            'upper_boundary': 5*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022,
       ...            'mean': 3*numpy.ones((2,)),
       ...            'sigma': 0.1}  # the global step-size may need to be tuned for better performance
       >>> lmcmaes = LMCMAES(problem, options)  # initialize the optimizer class
       >>> results = lmcmaes.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"LMCMAES: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       LMCMAES: 5000, 4.590739937885748e-16

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/yc7wyew4>`_ for more details.

    Attributes
    ----------
    c_c           : `float`
                    learning rate for evolution path update.
    c_s           : `float`
                    learning rate for population success rule.
    c_1           : `float`
                    learning rate for covariance matrix adaptation.
    d_s           : `float`
                    delay rate for population success rule.
    m             : `int`
                    number of direction vectors.
    mean          : `array_like`
                    initial (starting) point, aka mean of Gaussian search distribution.
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    n_parents     : `int`
                    number of parents, aka parental population size.
    n_steps       : `int`
                    target number of generations between vectors.
    sigma         : `float`
                    initial global step-size, aka mutation strength.
    z_star        : `float`
                    target success rate for population success rule.

    References
    ----------
    Loshchilov, I., 2014, July.
    A computationally efficient limited memory CMA-ES for large scale optimization.
    In Proceedings of Annual Conference on Genetic and Evolutionary Computation (pp. 397-404). ACM.
    https://dl.acm.org/doi/abs/10.1145/2576768.2598294

    See the official C++ version from Loshchilov:
    https://sites.google.com/site/lmcmaeses/
    """
    def __init__(self, problem, options):
        ES.__init__(self, problem, options)
        self.m = options.get('m', 4 + int(3*np.log(self.ndim_problem)))  # number of direction vectors
        self.n_steps = options.get('n_steps', self.m)  # target number of generations between vectors
        self.c_c = options.get('c_c', 1.0/self.m)  # learning rate for evolution path update
        self.c_1 = options.get('c_1', 1.0/(10.0*np.log(self.ndim_problem + 1.0)))
        self.c_s = options.get('c_s', 0.3)  # learning rate for population success rule (PSR)
        self.d_s = options.get('d_s', 1.0)  # damping parameter for PSR
        self.z_star = options.get('z_star', 0.25)  # target success rate for PSR
        self._a = np.sqrt(1.0 - self.c_1)
        self._c = 1.0/np.sqrt(1.0 - self.c_1)
        self._bd_1 = np.sqrt(1.0 - self.c_1)
        self._bd_2 = self.c_1/(1.0 - self.c_1)
        self._p_c_1 = 1.0 - self.c_c
        self._p_c_2 = None
        self._j = None
        self._l = None
        self._it = None
        self._rr = None  # for PSR

    def initialize(self, is_restart=False):
        mean = self._initialize_mean(is_restart)  # mean of Gaussian search distribution
        x = np.empty((self.n_individuals, self.ndim_problem))  # offspring population
        p_c = np.zeros((self.ndim_problem,))  # evolution path
        s = 0.0  # for PSR of global step-size adaptation
        vm = np.empty((self.m, self.ndim_problem))
        pm = np.empty((self.m, self.ndim_problem))
        b = np.empty((self.m,))
        d = np.empty((self.m,))
        y = np.empty((self.n_individuals,))  # fitness (no evaluation)
        self._p_c_2 = np.sqrt(self.c_c*(2.0 - self.c_c)*self._mu_eff)
        self._rr = np.arange(self.n_individuals*2, 0, -1) - 1
        self._j = [None]*self.m
        self._l = [None]*self.m
        self._it = 0
        return mean, x, p_c, s, vm, pm, b, d, y

    def _a_z(self, z=None, pm=None, vm=None, b=None):  # Algorithm 3 Az()
        x = np.copy(z)
        for t in range(self._it):
            x = self._a*x + b[self._j[t]]*np.dot(vm[self._j[t]], z)*pm[self._j[t]]
        return x

    def iterate(self, mean=None, x=None, pm=None, vm=None, y=None, b=None, args=None):
        sign, a_z = 1, np.empty((self.ndim_problem,))  # for mirrored sampling
        for k in range(self.n_individuals):
            if self._check_terminations():
                return x, y
            if sign == 1:
                z = self.rng_optimization.standard_normal((self.ndim_problem,))
                a_z = self._a_z(z, pm, vm, b)
            x[k] = mean + sign*self.sigma*a_z
            y[k] = self._evaluate_fitness(x[k], args)
            sign *= -1  # sampling in the opposite direction for mirrored sampling
        return x, y

    def _a_inv_z(self, v=None, vm=None, d=None, i=None):  # Algorithm 4 Ainvz()
        x = np.copy(v)
        for t in range(0, i):
            x = self._c*x - d[self._j[t]]*np.dot(vm[self._j[t]], x)*vm[self._j[t]]
        return x

    def _update_distribution(self, mean=None, x=None, p_c=None, s=None, vm=None,
                             pm=None, b=None, d=None, y=None, y_bak=None):
        mean_bak = np.dot(self._w, x[np.argsort(y)[:self.n_parents]])
        p_c = self._p_c_1*p_c + self._p_c_2*(mean_bak - mean)/self.sigma
        i_min = 1
        if self._n_generations < self.m:
            self._j[self._n_generations] = self._n_generations
        else:
            d_min = self._l[self._j[i_min]] - self._l[self._j[i_min - 1]]
            for j in range(2, self.m):
                d_cur = self._l[self._j[j]] - self._l[self._j[j - 1]]
                if d_cur < d_min:
                    d_min, i_min = d_cur, j
            # start from 0 if all pairwise distances exceed `self.n_steps`
            i_min = 0 if d_min >= self.n_steps else i_min
            # update indexes of evolution paths (`self._j[i_min]` is index of evolution path needed to delete)
            updated = self._j[i_min]
            for j in range(i_min, self.m - 1):
                self._j[j] = self._j[j + 1]
            self._j[self.m - 1] = updated
        self._it = np.minimum(self._n_generations + 1, self.m)
        self._l[self._j[self._it - 1]] = self._n_generations  # to update its generation
        pm[self._j[self._it - 1]] = p_c  # to add the latest evolution path
        # since `self._j[i_min]` is deleted, all vectors (from vm) depending on it need to be computed again
        for i in range(0 if i_min == 1 else i_min, self._it):
            vm[self._j[i]] = self._a_inv_z(pm[self._j[i]], vm, d, i)
            v_n = np.dot(vm[self._j[i]], vm[self._j[i]])
            bd_3 = np.sqrt(1.0 + self._bd_2*v_n)
            b[self._j[i]] = self._bd_1/v_n*(bd_3 - 1.0)
            d[self._j[i]] = 1.0/(self._bd_1*v_n)*(1.0 - 1.0/bd_3)
        if self._n_generations > 0:  # for population success rule (PSR)
            r = np.argsort(np.hstack((y, y_bak)))
            z_psr = np.sum(self._rr[r < self.n_individuals] - self._rr[r >= self.n_individuals])
            z_psr = z_psr/np.power(self.n_individuals, 2) - self.z_star
            s = (1.0 - self.c_s)*s + self.c_s*z_psr
            self.sigma *= np.exp(s/self.d_s)
        return mean_bak, p_c, s, vm, pm, b, d

    def restart_reinitialize(self, mean=None, x=None, p_c=None, s=None,
                             vm=None, pm=None, b=None, d=None, y=None):
        if self.is_restart and ES.restart_reinitialize(self, y):
            mean, x, p_c, s, vm, pm, b, d, y = self.initialize(True)
            self.d_s *= 2.0
        return mean, x, p_c, s, vm, pm, b, d, y

    def optimize(self, fitness_function=None, args=None):  # for all generations (iterations)
        fitness = ES.optimize(self, fitness_function)
        mean, x, p_c, s, vm, pm, b, d, y = self.initialize()
        while not self.termination_signal:
            y_bak = np.copy(y)
            # sample and evaluate offspring population
            x, y = self.iterate(mean, x, pm, vm, y, b, args)
            if self._check_terminations():
                break
            mean, p_c, s, vm, pm, b, d = self._update_distribution(
                mean, x, p_c, s, vm, pm, b, d, y, y_bak)
            self._print_verbose_info(fitness, y)
            self._n_generations += 1
            mean, x, p_c, s, vm, pm, b, d, y = self.restart_reinitialize(
                mean, x, p_c, s, vm, pm, b, d, y)
        results = self._collect(fitness, y, mean)
        results['p_c'] = p_c
        results['s'] = s
        return results



import numpy as np  # engine for numerical computing
from scipy.stats import multivariate_normal

from pypop7.optimizers.eda.eda import EDA
from pypop7.optimizers.eda.emna import EMNA


class REMNA(EMNA):
    """Reweighting Estimation of Multivariate Normal Algorithm (REMNA).

    Parameters
    ----------
    problem : dict
              problem arguments with the following common settings (`keys`):
                * 'fitness_function' - objective function to be **minimized** (`func`),
                * 'ndim_problem'     - number of dimensionality (`int`),
                * 'upper_boundary'   - upper boundary of search range (`array_like`),
                * 'lower_boundary'   - lower boundary of search range (`array_like`).
    options : dict
              optimizer options with the following common settings (`keys`):
                * 'max_function_evaluations' - maximum of function evaluations (`int`, default: `np.Inf`),
                * 'max_runtime'              - maximal runtime (`float`, default: `np.Inf`),
                * 'seed_rng'                 - seed for random number generation needed to be *explicitly* set (`int`);
              and with the following particular settings (`keys`):
                * 'n_individuals' - number of offspring, aka offspring population size (`int`, default: `200`),
                * 'n_parents'     - number of parents, aka parental population size (`int`, default:
                  `int(options['n_individuals']/2)`).

    Examples
    --------
    Use the optimizer `REMNA` to minimize the well-known test function
    `Rosenbrock <http://en.wikipedia.org/wiki/Rosenbrock_function>`_:

    .. code-block:: python
       :linenos:

       >>> import numpy  # engine for numerical computing
       >>> from pypop7.benchmarks.base_functions import rosenbrock  # function to be minimized
       >>> from pypop7.optimizers.eda.remna import REMNA
       >>> problem = {'fitness_function': rosenbrock,  # define problem arguments
       ...            'ndim_problem': 2,
       ...            'lower_boundary': -5.0*numpy.ones((2,)),
       ...            'upper_boundary': 5.0*numpy.ones((2,))}
       >>> options = {'max_function_evaluations': 5000,  # set optimizer options
       ...            'seed_rng': 2022}
       >>> remna = REMNA(problem, options)  # initialize the optimizer class
       >>> results = remna.optimize()  # run the optimization process
       >>> # return the number of function evaluations and best-so-far fitness
       >>> print(f"REMNA: {results['n_function_evaluations']}, {results['best_so_far_y']}")
       REMNA: 5000, 0.10884736030419488

    For its correctness checking of coding, refer to `this code-based repeatability report
    <https://tinyurl.com/2p8xksyy>`_ for more details.

    Attributes
    ----------
    n_individuals : `int`
                    number of offspring, aka offspring population size.
    n_parents     : `int`
                    number of parents, aka parental population size.

    References
    ----------
    Teytaud, F. and Teytaud, O., 2009, July.
    Why one must use reweighting in estimation of distribution algorithms.
    In Proceedings of ACM Annual Conference on Genetic and Evolutionary Computation (pp. 453-460)
    https://doi.org/10.1145/1569901.1569964
    """
    def __init__(self, problem, options):
        EMNA.__init__(self, problem, options)

    def initialize(self, args=None):
        mean, cov = (self.initial_upper_boundary + self.initial_lower_boundary)/2.0, 0.1*np.eye(self.ndim_problem)
        x = self.rng_optimization.multivariate_normal(mean, cov, size=(self.n_individuals,))  # population
        y = np.empty((self.n_individuals,))  # fitness
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y, mean, cov

    def iterate(self, x=None, y=None, mean=None, cov=None, args=None):
        order = np.argsort(y)[:self.n_parents]
        try:
            m = multivariate_normal(mean=mean, cov=cov)
        except Exception:
            m = multivariate_normal(mean=mean, cov=cov + 1e-16*np.eye(self.ndim_problem))
        w = 1.0/m.pdf(x[order]).reshape(-1, 1)
        x[order] += (x[order] - mean)*w/np.sum(w)
        mean = np.mean(x[order], axis=0)
        cov = np.cov(np.transpose(x[order]))
        for i in range(self.n_individuals):
            if self._check_terminations():
                break
            x[i] = self.rng_optimization.multivariate_normal(mean, cov)
            y[i] = self._evaluate_fitness(x[i], args)
        return x, y, mean, cov

    def optimize(self, fitness_function=None, args=None):
        fitness = EDA.optimize(self, fitness_function)
        x, y, mean, cov = self.initialize(args)
        while not self._check_terminations():
            self._print_verbose_info(fitness, y)
            x, y, mean, cov = self.iterate(x, y, mean, cov, args)
            self._n_generations += 1
        return self._collect(fitness, y)



"""Repeat the following paper for `MRAS`:
    Hu, J., Fu, M.C. and Marcus, S.I., 2007.
    A model reference adaptive search method for global optimization.
    Operations Research, 55(3), pp.549-568.
    https://pubsonline.informs.org/doi/abs/10.1287/opre.1060.0367

    Luckily our Python code could repeat the data generated by the original paper *nearly well*.
    Therefore, we argue that its repeatability could be **well-documented**.
"""
import time

import numpy as np

from pypop7.benchmarks.base_functions import griewank
from pypop7.optimizers.cem.mras import MRAS as Solver


if __name__ == '__main__':
    start_run = time.time()
    ndim_problem = 20

    problem = {'fitness_function': griewank,
               'ndim_problem': ndim_problem,
               'lower_boundary': -50*np.ones((ndim_problem,)),
               'upper_boundary': 50*np.ones((ndim_problem,))}
    options = {'max_function_evaluations': 400000,
               'seed_rng': 0,
               'sigma': np.sqrt(500),
               'verbose': 20,
               'saving_fitness': 2000}
    solver = Solver(problem, options)
    results = solver.optimize()
    print(results)  # 3.87685168e-01 vs 4e7e−03 (from the original paper)
    print('*** Runtime: {:7.5e}'.format(time.time() - start_run))



from pypop7.benchmarks.utils import plot_contour
from pypop7.benchmarks.base_functions import ackley, rastrigin
from pypop7.benchmarks.rotated_functions import generate_rotation_matrix, ellipsoid


def cd(x):  # from https://arxiv.org/pdf/1610.00040v1.pdf
    return 7*(x[0]**2) + 6*x[0]*x[1] + 8*(x[1]**2)


if __name__ == '__main__':
    # plot multi-modality
    plot_contour(rastrigin, [-10, 10], [-10, 10])
    # plot non-convexity
    plot_contour(ackley, [-10, 10], [-10, 10])
    # plot non-separability
    plot_contour(cd, [-10, 10], [-10, 10])
    # plot ill-condition
    generate_rotation_matrix(ellipsoid, 2, 72)
    plot_contour(ellipsoid, [-10, 10], [-10, 10],
                 levels=[1e0, 1e3, 1e6, 5e6, 1e7, 5e7, 1e8, 5e8, 1e9])



"""Repeat the following paper for `ESA`:
    Siarry, P., Berthiau, G., Durdin, F. and Haussy, J., 1997.
    Enhanced simulated annealing for globally minimizing functions of many-continuous variables.
    ACM Transactions on Mathematical Software, 23(2), pp.209-228.
    https://dl.acm.org/doi/abs/10.1145/264029.264043

    Since several implementation details were not very clear in the original paper, it is hard to
      repeat the original experiments *perfectly*, according to our experiences.
    Our current Python implementation could generate the *relatively close but not perfect* result on the
    following test function.
    We still expect to improve it, in order to match the original paper as close as possible in the future
      (if possible).

    The fuzzy details from the original paper are presented below, which can cause different implementations
      with different performances:
    1. Initial FOBJ variation average, DGYINI, obtained from (typically) 50 uphill moves performed
      before beginning SA minimization itself.
    2. Step 2: Space Partitioning.
    3. MSOTST = 0 (this should be a typo in the original paper).
    4. Other systematic experiments are still in progress in order to obtain, if possible, a rule for
      the p-optimal choice, adapted to FOBJ-specific characteristics.
"""
import numpy as np

from pypop7.benchmarks.base_functions import rosenbrock
from pypop7.optimizers.sa.esa import ESA


if __name__ == '__main__':
    problem = {'fitness_function': rosenbrock,
               'ndim_problem': 50,
               'upper_boundary': 10*np.ones((50,)),
               'lower_boundary': -5*np.ones((50,))}
    options = {'max_function_evaluations': 78224,
               'seed_rng': 1,  # undefined in the original paper
               'p': 1,
               'saving_fitness': 1000,
               'verbose': 10000}
    esa = ESA(problem, options)
    results = esa.optimize()
    # FOBJ initial-value average: 6472239 (from 50 samples) vs 5e5 (from the original paper)
    # our implementation starts from a worse starting point than the original paper.
    print(results)  # 49.13408124446869 vs 8.8 (from the original paper)
